import abc
import logging

from copy import deepcopy
from typing import Any, cast, Dict, Iterator, List, MutableMapping, Optional, Tuple, Type, Union

import yaml

from ..collections import CaseInsensitiveDict, CaseInsensitiveSet
from ..exceptions import PatroniException
from ..utils import parse_bool, parse_int, parse_real
from .available_parameters import get_validator_files, PathLikeObj

logger = logging.getLogger(__name__)


class _Transformable(abc.ABC):

    def __init__(self, version_from: int, version_till: Optional[int] = None) -> None:
        self.__version_from = version_from
        self.__version_till = version_till

    @classmethod
    def get_subclasses(cls) -> Iterator[Type['_Transformable']]:
        """Recursively get all subclasses of :class:`_Transformable`.

        :yields: each subclass of :class:`_Transformable`.
        """
        for subclass in cls.__subclasses__():
            yield from subclass.get_subclasses()
            yield subclass

    @property
    def version_from(self) -> int:
        return self.__version_from

    @property
    def version_till(self) -> Optional[int]:
        return self.__version_till

    @abc.abstractmethod
    def transform(self, name: str, value: Any) -> Optional[Any]:
        """Verify that provided value is valid.

        :param name: GUC's name
        :param value: GUC's value
        :returns: the value (sometimes clamped) or ``None`` if the value isn't valid
        """


class Bool(_Transformable):

    def transform(self, name: str, value: Any) -> Optional[Any]:
        if parse_bool(value) is not None:
            return value
        logger.warning('Removing bool parameter=%s from the config due to the invalid value=%s', name, value)


class Number(_Transformable):

    def __init__(self, *, version_from: int, version_till: Optional[int] = None, min_val: Union[int, float],
                 max_val: Union[int, float], unit: Optional[str] = None) -> None:
        super(Number, self).__init__(version_from, version_till)
        self.__min_val = min_val
        self.__max_val = max_val
        self.__unit = unit

    @property
    def min_val(self) -> Union[int, float]:
        return self.__min_val

    @property
    def max_val(self) -> Union[int, float]:
        return self.__max_val

    @property
    def unit(self) -> Optional[str]:
        return self.__unit

    @staticmethod
    @abc.abstractmethod
    def parse(value: Any, unit: Optional[str]) -> Optional[Any]:
        """Convert provided value to unit."""

    def transform(self, name: str, value: Any) -> Union[int, float, None]:
        num_value = self.parse(value, self.unit)
        if num_value is not None:
            if num_value < self.min_val:
                logger.warning('Value=%s of parameter=%s is too low, increasing to %s%s',
                               value, name, self.min_val, self.unit or '')
                return self.min_val
            if num_value > self.max_val:
                logger.warning('Value=%s of parameter=%s is too big, decreasing to %s%s',
                               value, name, self.max_val, self.unit or '')
                return self.max_val
            return value
        logger.warning('Removing %s parameter=%s from the config due to the invalid value=%s',
                       self.__class__.__name__.lower(), name, value)


class Integer(Number):

    @staticmethod
    def parse(value: Any, unit: Optional[str]) -> Optional[int]:
        return parse_int(value, unit)


class Real(Number):

    @staticmethod
    def parse(value: Any, unit: Optional[str]) -> Optional[float]:
        return parse_real(value, unit)


class Enum(_Transformable):

    def __init__(self, *, version_from: int, version_till: Optional[int] = None,
                 possible_values: Tuple[str, ...]) -> None:
        super(Enum, self).__init__(version_from, version_till)
        self.__possible_values = possible_values

    @property
    def possible_values(self) -> Tuple[str, ...]:
        return self.__possible_values

    def transform(self, name: str, value: Optional[Any]) -> Optional[Any]:
        if str(value).lower() in self.possible_values:
            return value
        logger.warning('Removing enum parameter=%s from the config due to the invalid value=%s', name, value)


class EnumBool(Enum):

    def transform(self, name: str, value: Optional[Any]) -> Optional[Any]:
        if parse_bool(value) is not None:
            return value
        return super(EnumBool, self).transform(name, value)


class String(_Transformable):

    def transform(self, name: str, value: Optional[Any]) -> Optional[Any]:
        return value


# Format:
#  key  - parameter name
#  value - variable length tuple of `_Transformable` objects. Each object in the tuple represents a different
#          validation of the GUC across postgres versions. If a GUC validation has never changed over time, then it will
#          have a single object in the tuple. For example, `password_encryption` used to be a boolean GUC up to Postgres
#          10, at which point it started being an enum. In that case the value of `password_encryption` would be a tuple
#          of 2 `_Transformable` objects (`Bool` and `Enum`, respectively), each one reprensenting a different
#          validation rule.
parameters = CaseInsensitiveDict()
recovery_parameters = CaseInsensitiveDict()


class ValidatorFactoryNoType(PatroniException):
    """Raised when a validator spec misses a type."""


class ValidatorFactoryInvalidType(PatroniException):
    """Raised when a validator spec contains an invalid type."""


class ValidatorFactoryInvalidSpec(PatroniException):
    """Raised when a validator spec contains an invalid set of attributes."""


class ValidatorFactory:
    """Factory class used to build Patroni validator objects based on the given specs."""

    TYPES: Dict[str, Type[_Transformable]] = {cls.__name__: cls for cls in _Transformable.get_subclasses()}

    def __new__(cls, validator: Dict[str, Any]) -> _Transformable:
        """Parse a given Postgres GUC *validator* into the corresponding Patroni validator object.

        :param validator: a validator spec for a given parameter. It usually comes from a parsed YAML file.

        :returns: the Patroni validator object that corresponds to the specification found in *validator*.

        :raises:
            :class:`ValidatorFactoryNoType`: if *validator* contains no ``type`` key.
            :class:`ValidatorFactoryInvalidType`: if ``type`` key from *validator* contains an invalid value.
            :class:`ValidatorFactoryInvalidSpec`: if *validator* contains an invalid set of attributes for the given
                ``type``.

        :Example:

            If a given validator was defined as follows in the YAML file:

            ```yaml
            - type: String
            version_from: 90300
            version_till: null
            ```

            Then this method would receive *validator* as:

            ```python
            {
                'type': 'String',
                'version_from': 90300,
                'version_till': None
            }
            ```

            And this method would return a :class:`String`:

            ```python
            String(90300, None)
            ```
        """
        validator = deepcopy(validator)
        try:
            type_ = validator.pop('type')
        except KeyError as exc:
            raise ValidatorFactoryNoType('Validator contains no type.') from exc

        if type_ not in cls.TYPES:
            raise ValidatorFactoryInvalidType(f'Unexpected validator type: `{type_}`.')

        for key, value in validator.items():
            # :func:`_transform_parameter_value` expects :class:`tuple` instead of :class:`list`
            if isinstance(value, list):
                validator[key] = tuple(cast(List[Any], value))

        try:
            return cls.TYPES[type_](**validator)
        except Exception as exc:
            raise ValidatorFactoryInvalidSpec(
                f'Failed to parse `{type_}` validator (`{validator}`): `{str(exc)}`.') from exc


def _get_postgres_guc_validators(config: Dict[str, Any], parameter: str) -> Tuple[_Transformable, ...]:
    """Get all validators of *parameter* from *config*.

    Loop over all validators specs of *parameter* and return them parsed as Patroni validators.

    :param config: Python object corresponding to an YAML file, with values of either ``parameters`` or
        ``recovery_parameters`` key.
    :param parameter: name of the parameter found under *config* which validators should be parsed and returned.

    :rtype: yields any exception that is faced while parsing a validator spec into a Patroni validator object.
    """
    validators: List[_Transformable] = []
    for validator_spec in config.get(parameter, []):
        try:
            validator = ValidatorFactory(validator_spec)
            validators.append(validator)
        except (ValidatorFactoryNoType, ValidatorFactoryInvalidType, ValidatorFactoryInvalidSpec) as exc:
            logger.warning('Faced an issue while parsing a validator for parameter `%s`: `%r`', parameter, exc)

    return tuple(validators)


class InvalidGucValidatorsFile(PatroniException):
    """Raised when reading or parsing of a YAML file faces an issue."""


def _read_postgres_gucs_validators_file(file: PathLikeObj) -> Dict[str, Any]:
    """Read an YAML file and return the corresponding Python object.

    :param file: path-like object to read from. It is expected to be encoded with ``UTF-8``, and to be a YAML document.

    :returns: the YAML content parsed into a Python object. If any issue is faced while reading/parsing the file, then
        return ``None``.

    :raises:
        :class:`InvalidGucValidatorsFile`: if faces an issue while reading or parsing *file*.
    """
    try:
        with file.open(encoding='UTF-8') as stream:
            return yaml.safe_load(stream)
    except Exception as exc:
        raise InvalidGucValidatorsFile(
            f'Unexpected issue while reading parameters file `{file}`: `{str(exc)}`.') from exc


def _load_postgres_gucs_validators() -> None:
    """Load all Postgres GUC validators from YAML files.

    Recursively walk through ``available_parameters`` directory and load validators of each found YAML file into
    ``parameters`` and/or ``recovery_parameters`` variables.

    Walk through directories in top-down fashion and for each of them:
        * Sort files by name;
        * Load validators from YAML files that were found.

    Any problem faced while reading or parsing files will be logged as a ``WARNING`` by the child function, and the
    corresponding file or validator will be ignored.

    By default, Patroni only ships the file ``0_postgres.yml``, which contains Community Postgres GUCs validators, but
    that behavior can be extended. For example: if a vendor wants to add GUC validators to Patroni for covering a custom
    Postgres build, then they can create their custom YAML files under ``available_parameters`` directory.

    Each YAML file may contain either or both of these root attributes, here called sections:
        * ``parameters``: general GUCs that would be written to ``postgresql.conf``;
        * ``recovery_parameters``: recovery related GUCs that would be written to ``recovery.conf`` (Patroni later
            writes them to ``postgresql.conf`` if running PG 12 and above).

    Then, each of these sections, if specified, may contain one or more attributes with the following structure:

        * key: the name of a GUC;
        * value: a list of validators. Each item in the list must contain a ``type`` attribute, which must be one among:

            * ``Bool``; or
            * ``Integer``; or
            * ``Real``; or
            * ``Enum``; or
            * ``EnumBool``; or
            * ``String``.

            Besides the ``type`` attribute, it should also contain all the required attributes as per the corresponding
            class in this module.

    .. seealso::

        * :class:`Bool`;
        * :class:`Integer`;
        * :class:`Real`;
        * :class:`Enum`;
        * :class:`EnumBool`;
        * :class:`String`.

    :Example:

        This is a sample content for an YAML file based on Postgres GUCs, showing each of the supported types and
        sections:

            .. code-block:: yaml

                parameters:
                  archive_command:
                  - type: String
                    version_from: 90300
                    version_till: null
                  archive_mode:
                  - type: Bool
                    version_from: 90300
                    version_till: 90500
                  - type: EnumBool
                    version_from: 90500
                    version_till: null
                    possible_values:
                    - always
                  archive_timeout:
                  - type: Integer
                    version_from: 90300
                    version_till: null
                    min_val: 0
                    max_val: 1073741823
                    unit: s
                  autovacuum_vacuum_cost_delay:
                  - type: Integer
                    version_from: 90300
                    version_till: 120000
                    min_val: -1
                    max_val: 100
                    unit: ms
                  - type: Real
                    version_from: 120000
                    version_till: null
                    min_val: -1
                    max_val: 100
                    unit: ms
                  client_min_messages:
                  - type: Enum
                    version_from: 90300
                    version_till: null
                    possible_values:
                    - debug5
                    - debug4
                    - debug3
                    - debug2
                    - debug1
                    - log
                    - notice
                    - warning
                    - error
                recovery_parameters:
                  archive_cleanup_command:
                  - type: String
                    version_from: 90300
                    version_till: null

    """
    for file in get_validator_files():
        try:
            config: Dict[str, Any] = _read_postgres_gucs_validators_file(file)
        except InvalidGucValidatorsFile as exc:
            logger.warning(str(exc))
            continue

        logger.debug('Parsing validators from file `%s`.', file)

        mapping = {
            'parameters': parameters,
            'recovery_parameters': recovery_parameters,
        }

        for section in ['parameters', 'recovery_parameters']:
            section_var = mapping[section]

            config_section = config.get(section, {})
            for parameter in config_section.keys():
                section_var[parameter] = _get_postgres_guc_validators(config_section, parameter)


_load_postgres_gucs_validators()


def _transform_parameter_value(validators: MutableMapping[str, Tuple[_Transformable, ...]],
                               version: int, name: str, value: Any,
                               available_gucs: CaseInsensitiveSet) -> Optional[Any]:
    """Validate *value* of GUC *name* for Postgres *version* using defined *validators* and *available_gucs*.

    :param validators: a dictionary of all GUCs across all Postgres versions. Each key is the name of a Postgres GUC,
        and the corresponding value is a variable length tuple of :class:`_Transformable`. Each item is a validation
        rule for the GUC for a given range of Postgres versions. Should either contain recovery GUCs or general GUCs,
        not both.
    :param version: Postgres version to validate the GUC against.
    :param name: name of the Postgres GUC.
    :param value: value of the Postgres GUC.
    :param available_gucs: a set of all GUCs available in Postgres *version*. Each item is the name of a Postgres
        GUC. Used to avoid ignoring GUC *name* if it does not have a validator in *validators*, but is a valid GUC
        in Postgres *version*.

    :returns: the return value may be one among:

        * *value* transformed to the expected format for GUC *name* in Postgres *version*, if *name* has a validator
          in *validators* for the corresponding Postgres *version*; or
        * ``None`` if *name* does not have a validator in *validators* and is not present in *available_gucs*.
    """
    for validator in validators.get(name, ()) or ():
        if version >= validator.version_from and\
                (validator.version_till is None or version < validator.version_till):
            return validator.transform(name, value)
    # Ideally we should have a validator in *validators*. However, if none is available, we will not discard a
    # setting that exists in Postgres *version*, but rather allow the value with no validation.
    if name in available_gucs:
        return value
    logger.warning('Removing unexpected parameter=%s value=%s from the config', name, value)


def transform_postgresql_parameter_value(version: int, name: str, value: Any,
                                         available_gucs: CaseInsensitiveSet) -> Optional[Any]:
    """Validate *value* of GUC *name* for Postgres *version* using ``parameters`` and *available_gucs*.

    :param version: Postgres version to validate the GUC against.
    :param name: name of the Postgres GUC.
    :param value: value of the Postgres GUC.
    :param available_gucs: a set of all GUCs available in Postgres *version*. Each item is the name of a Postgres
        GUC. Used to avoid ignoring GUC *name* if it does not have a validator in ``parameters``, but is a valid GUC in
        Postgres *version*.

    :returns: The return value may be one among:

        * The original *value* if *name* seems to be an extension GUC (contains a period '.'); or
        * ``None`` if **name** is a recovery GUC; or
        * *value* transformed to the expected format for GUC *name* in Postgres *version* using validators defined in
            ``parameters``. Can also return ``None``. See :func:`_transform_parameter_value`.
    """
    if '.' in name and name not in parameters:
        # likely an extension GUC, so just return as it is. Otherwise, if `name` is in `parameters`, it's likely a
        # namespaced GUC from a custom Postgres build, so we treat that over the usual validation means.
        return value
    if name in recovery_parameters:
        return None
    return _transform_parameter_value(parameters, version, name, value, available_gucs)


def transform_recovery_parameter_value(version: int, name: str, value: Any,
                                       available_gucs: CaseInsensitiveSet) -> Optional[Any]:
    """Validate *value* of GUC *name* for Postgres *version* using ``recovery_parameters`` and *available_gucs*.

    :param version: Postgres version to validate the recovery GUC against.
    :param name: name of the Postgres recovery GUC.
    :param value: value of the Postgres recovery GUC.
    :param available_gucs: a set of all GUCs available in Postgres *version*. Each item is the name of a Postgres
        GUC. Used to avoid ignoring GUC *name* if it does not have a validator in ``parameters``, but is a valid GUC in
        Postgres *version*.

    :returns: *value* transformed to the expected format for recovery GUC *name* in Postgres *version* using validators
        defined in ``recovery_parameters``. It can also return ``None``. See :func:`_transform_parameter_value`.
    """
    # Recovery settings are not present in ``postgres --describe-config`` output of Postgres <= 11. In that case we
    # just pass down the list of settings defined in Patroni validators so :func:`_transform_parameter_value` will not
    # discard the recovery GUCs when running Postgres <= 11.
    # NOTE: At the moment this change was done Postgres 11 was almost EOL, and had been likely extensively used with
    # Patroni, so we should be able to rely solely on Patroni validators as the source of truth.
    return _transform_parameter_value(
        recovery_parameters, version, name, value,
        available_gucs if version >= 120000 else CaseInsensitiveSet(recovery_parameters.keys()))



"""Replication slot handling.

Provides classes for the creation, monitoring, management and synchronisation of PostgreSQL replication slots.
"""

import logging
import os
import shutil

from collections import defaultdict
from contextlib import contextmanager
from threading import Condition, Thread
from typing import Any, Dict, Generator, List, Optional, Tuple, TYPE_CHECKING, Union

from .. import global_config
from ..dcs import Cluster, Leader
from ..file_perm import pg_perm
from ..psycopg import OperationalError
from ..tags import Tags
from .connection import get_connection_cursor
from .misc import format_lsn, fsync_dir, PostgresqlRole

if TYPE_CHECKING:  # pragma: no cover
    from psycopg import Cursor
    from psycopg2 import cursor

    from . import Postgresql

logger = logging.getLogger(__name__)


def compare_slots(s1: Dict[str, Any], s2: Dict[str, Any], dbid: str = 'database') -> bool:
    """Compare 2 replication slot objects for equality.

    ..note ::
        If the first argument is a ``physical`` replication slot then only the `type` of the second slot is compared.
        If the first argument is another ``type`` (e.g. ``logical``) then *dbid* and ``plugin`` are compared.

    :param s1: First slot dictionary to be compared.
    :param s2: Second slot dictionary to be compared.
    :param dbid: Optional attribute to be compared when comparing ``logical`` replication slots.

    :return: ``True`` if the slot ``type`` of *s1* and *s2* is matches, and the ``type`` of *s1* is ``physical``,
             OR the ``types`` match AND the *dbid* and ``plugin`` attributes are equal.

    """
    return (s1['type'] == s2['type']
            and (s1['type'] == 'physical'
                 or s1.get(dbid) == s2.get(dbid)
                 and s1['plugin'] == s2['plugin']))


class SlotsAdvanceThread(Thread):
    """Daemon process :class:``Thread`` object for advancing logical replication slots on replicas.

    This ensures that slot advancing queries sent to postgres do not block the main loop.
    """

    def __init__(self, slots_handler: 'SlotsHandler') -> None:
        """Create and start a new thread for handling slot advance queries.

        :param slots_handler: The calling class instance for reference to slot information attributes.
        """
        super().__init__()
        self.daemon = True
        self._slots_handler = slots_handler

        # _copy_slots and _failed are used to asynchronously give some feedback to the main thread
        self._copy_slots: List[str] = []
        self._failed = False
        # {'dbname1': {'slot1': 100, 'slot2': 100}, 'dbname2': {'slot3': 100}}
        self._scheduled: Dict[str, Dict[str, int]] = defaultdict(dict)
        self._condition = Condition()  # protect self._scheduled from concurrent access and to wakeup the run() method

        self.start()

    def sync_slot(self, cur: Union['cursor', 'Cursor[Any]'], database: str, slot: str, lsn: int) -> None:
        """Execute a ``pg_replication_slot_advance`` query and store success for scheduled synchronisation task.

        :param cur: database connection cursor.
        :param database: name of the database associated with the slot.
        :param slot: name of the slot to be synchronised.
        :param lsn: last known LSN position
        """
        failed = copy = False
        try:
            cur.execute("SELECT pg_catalog.pg_replication_slot_advance(%s, %s)", (slot, format_lsn(lsn)))
        except Exception as e:
            logger.error("Failed to advance logical replication slot '%s': %r", slot, e)
            failed = True
            # WAL file is gone or slot is invalidated
            copy = isinstance(e, OperationalError) and e.diag.sqlstate in ('58P01', '55000')
        with self._condition:
            if self._scheduled and failed:
                if copy and slot not in self._copy_slots:
                    self._copy_slots.append(slot)
                self._failed = True

            new_lsn = self._scheduled.get(database, {}).get(slot, 0)
            # remove slot from the self._scheduled structure if it is to be copied or if it wasn't changed
            if copy or (new_lsn == lsn and database in self._scheduled):
                self._scheduled[database].pop(slot)
                if not self._scheduled[database]:
                    self._scheduled.pop(database)

    def sync_slots_in_database(self, database: str, slots: List[str]) -> None:
        """Synchronise slots for a single database.

        :param database: name of the database.
        :param slots: list of slot names to synchronise.
        """
        with self._slots_handler.get_local_connection_cursor(dbname=database, options='-c statement_timeout=0') as cur:
            for slot in slots:
                with self._condition:
                    lsn = self._scheduled.get(database, {}).get(slot, 0)
                if lsn:
                    self.sync_slot(cur, database, slot, lsn)

    def sync_slots(self) -> None:
        """Synchronise slots for all scheduled databases."""
        with self._condition:
            databases = list(self._scheduled.keys())
        for database in databases:
            with self._condition:
                slots = list(self._scheduled.get(database, {}).keys())
            if slots:
                try:
                    self.sync_slots_in_database(database, slots)
                except Exception as e:
                    logger.error('Failed to advance replication slots in database %s: %r', database, e)

    def run(self) -> None:
        """Thread main loop entrypoint.

        .. note::
            Thread will wait until a sync is scheduled from outside, normally triggered during the HA loop or a wakeup
            call.
        """
        while True:
            with self._condition:
                if not self._scheduled:
                    self._condition.wait()

            self.sync_slots()

    def schedule(self, advance_slots: Dict[str, Dict[str, int]]) -> Tuple[bool, List[str]]:
        """Trigger a synchronisation of slots.

        This is the main entrypoint for Patroni HA loop wakeup call.

        :param advance_slots: dictionary containing slots that need to be advanced

        :return: tuple of failure status and a list of slots to be copied
        """
        with self._condition:
            for database, values in advance_slots.items():
                for name, value in values.items():
                    # Don't schedule sync for slots that just failed to be advanced and scheduled to be copied
                    if name not in self._copy_slots:
                        self._scheduled[database][name] = value
            ret = (self._failed, self._copy_slots)
            self._copy_slots = []
            self._failed = False
            self._condition.notify()

        return ret

    def clean(self) -> None:
        """Reset state of the daemon."""
        with self._condition:
            self._scheduled.clear()
            self._failed = False
            self._copy_slots = []


class SlotsHandler:
    """Handler for managing and storing information on replication slots in PostgreSQL.

    :ivar pg_replslot_dir: system location path of the PostgreSQL replication slots.
    :ivar _logical_slots_processing_queue: yet to be processed logical replication slots on the primary
    """

    def __init__(self, postgresql: 'Postgresql') -> None:
        """Create an instance with storage attributes for replication slots and schedule the first synchronisation.

        :param postgresql: Calling class instance providing interface to PostgreSQL.
        """
        self._force_readiness_check = False
        self._schedule_load_slots = False
        self._postgresql = postgresql
        self._advance = SlotsAdvanceThread(self)
        self._replication_slots: Dict[str, Dict[str, Any]] = {}  # already existing replication slots
        self._logical_slots_processing_queue: Dict[str, Optional[int]] = {}
        self.pg_replslot_dir = os.path.join(self._postgresql.data_dir, 'pg_replslot')
        self.schedule()

    def _query(self, sql: str, *params: Any) -> List[Tuple[Any, ...]]:
        """Helper method for :meth:`Postgresql.query`.

        :param sql: SQL statement to execute.
        :param params: parameters to pass through to :meth:`Postgresql.query`.

        :returns: query response.
        """
        return self._postgresql.query(sql, *params, retry=False)

    def copy_slot_items(self, src: Dict[str, Any], dst: Dict[str, Any], is_logical: bool = True) -> None:
        """Select values from *src* slots dictionary to update in *dst* slots dictionary.

        :param src: source dictionary that *keys* will be looked up from.
        :param dst: destination dictionary to be updated.
        :param is_logical: whether the slot type is logical.
        """
        if is_logical:
            keys = ('datoid', 'catalog_xmin', 'confirmed_flush_lsn')
        else:
            keys = ('restart_lsn', 'xmin')
        if self._postgresql.major_version >= 130000:
            keys += ('wal_status',)
        dst.update({key: src[key] for key in keys if key in src})

    def process_permanent_slots(self, slots: List[Dict[str, Any]]) -> Dict[str, int]:
        """Process replication slot information from the host and prepare information used in subsequent cluster tasks.

        .. note::
            This methods solves three problems.

            The ``cluster_info_query`` from :class:``Postgresql`` is executed every HA loop and returns information
            about all replication slots that exists on the current host.

            Based on this information perform the following actions:

            1. For the primary we want to expose to DCS permanent logical slots, therefore build (and return) a dict
               that maps permanent logical slot names to ``confirmed_flush_lsn``.
            2. detect if one of the previously known permanent slots is missing and schedule resync.
            3. Update the local cache with the fresh ``catalog_xmin`` and ``confirmed_flush_lsn`` for every known slot.

           This info is used when performing the check of logical slot readiness on standbys.

        :param slots: replication slot information that exists on the current host.

        :return: dictionary of logical slot names to ``confirmed_flush_lsn``.
        """
        ret: Dict[str, int] = {}

        slots_dict: Dict[str, Dict[str, Any]] = {slot['slot_name']: slot for slot in slots or []}
        for name, value in slots_dict.items():
            if name in self._replication_slots:
                if compare_slots(value, self._replication_slots[name], 'datoid'):
                    if value['type'] == 'logical':
                        ret[name] = value['confirmed_flush_lsn']
                        self.copy_slot_items(value, self._replication_slots[name])
                    else:
                        ret[name] = value['restart_lsn']
                        self.copy_slot_items(value, self._replication_slots[name], is_logical=False)
                else:
                    self._schedule_load_slots = True

        # It could happen that the slot was deleted in the background, we want to detect this case
        if any(name not in slots_dict for name in self._replication_slots.keys()):
            self._schedule_load_slots = True

        return ret

    def pg_replication_slots_query(self, use_function: bool = False) -> str:
        """Get query text for retrieveing slots infromation.

        Query retrieves replication slot's ``name``, ``type``, ``plugin``, ``database`` and ``datoid``.
        If PostgreSQL version is 10 or newer also retrieves ``catalog_xmin`` and ``confirmed_flush_lsn``.

        :param use_function: whether to use ``pg_get_replication_slots()`` function or ``pg_replication_slots`` view
                             in the query.

        :return: string containing SQL query text.
        """
        pg_wal_lsn_diff = f"pg_catalog.pg_{self._postgresql.wal_name}_{self._postgresql.lsn_name}_diff"

        extra = f", catalog_xmin, {pg_wal_lsn_diff}(confirmed_flush_lsn, '0/0')::bigint AS confirmed_flush_lsn" \
            if self._postgresql.major_version >= 100000 else ""
        extra += ", wal_status" if self._postgresql.major_version >= 130000 else ""

        filter_columns = ["NOT temporary"] if self._postgresql.major_version >= 100000 else []
        filter_columns += ["NOT failover"] if self._postgresql.major_version >= 170000 else []

        where_filter = ' AND '.join(filter_columns)
        where_condition = f' WHERE {where_filter}' if where_filter else ''

        database = '' if use_function else ', database'
        pg_replication_slots_obj = "pg_get_replication_slots()" if use_function else 'pg_replication_slots'

        return "SELECT slot_name, slot_type AS type, xmin, " \
            f"{pg_wal_lsn_diff}(restart_lsn, '0/0')::bigint AS restart_lsn, plugin{database}, " \
            f"datoid::bigint{extra} FROM pg_catalog.{pg_replication_slots_obj}{where_condition}"

    def load_replication_slots(self) -> None:
        """Query replication slot information from the database and store it for processing by other tasks.

        .. note::
            Only supported from PostgreSQL version 9.4 onwards.

        When using logical slots, store information separately for slot synchronisation on replica nodes.
        """
        if self._postgresql.major_version >= 90400 and self._schedule_load_slots:
            replication_slots: Dict[str, Dict[str, Any]] = {}
            for r in self._query(self.pg_replication_slots_query()):
                value = {'type': r[1]}
                if r[1] == 'logical':
                    value.update(plugin=r[4], database=r[5], datoid=r[6])
                    if self._postgresql.major_version >= 100000:
                        value.update(catalog_xmin=r[7], confirmed_flush_lsn=r[8])
                else:
                    value.update(xmin=r[2], restart_lsn=r[3])
                if self._postgresql.major_version >= 130000:
                    value.update(wal_status=r[9])
                replication_slots[r[0]] = value
            self._replication_slots = replication_slots
            self._schedule_load_slots = False
            if self._force_readiness_check:
                self._logical_slots_processing_queue = {n: None for n, v in replication_slots.items()
                                                        if v['type'] == 'logical'}
                self._force_readiness_check = False

    def ignore_replication_slot(self, cluster: Cluster, name: str) -> bool:
        """Check if slot *name* should not be managed by Patroni.

        :param cluster: cluster state information object.
        :param name: name of the slot to ignore

        :returns: ``True`` if slot *name* matches any slot specified in ``ignore_slots`` configuration,
                 otherwise will pass through and return result of :meth:`AbstractMPPHandler.ignore_replication_slot`.
        """
        slot = self._replication_slots[name]
        if cluster.config:
            for matcher in global_config.ignore_slots_matchers:
                if (
                        (matcher.get("name") is None or matcher["name"] == name)
                        and all(not matcher.get(a) or matcher[a] == slot.get(a)
                                for a in ('database', 'plugin', 'type'))
                ):
                    return True
        return self._postgresql.mpp_handler.ignore_replication_slot(slot)

    def drop_replication_slot(self, name: str) -> Tuple[bool, bool]:
        """Drop a named slot from Postgres.

        :param name: name of the slot to be dropped.

        :returns: a tuple of ``active`` and ``dropped``. ``active`` is ``True`` if the slot is active,
                  ``dropped`` is ``True`` if the slot was successfully dropped. If the slot was not found return
                  ``False`` for both.
        """
        rows = self._query(('WITH slots AS (SELECT slot_name, active'
                            ' FROM pg_catalog.pg_replication_slots WHERE slot_name = %s),'
                            ' dropped AS (SELECT pg_catalog.pg_drop_replication_slot(slot_name),'
                            ' true AS dropped FROM slots WHERE not active) '
                            'SELECT active, COALESCE(dropped, false) FROM slots'
                            ' FULL OUTER JOIN dropped ON true'), name)
        return (rows[0][0], rows[0][1]) if rows else (False, False)

    def _drop_replication_slot(self, name: str) -> None:
        """Drop replication slot by name.

        .. note::
            If not able to drop the slot, it will log a message and set the flag to reload slots.

        :param name: name of the slot to be dropped.
        """
        active, dropped = self.drop_replication_slot(name)
        if dropped:
            logger.info("Dropped replication slot '%s'", name)
            if name in self._replication_slots:
                del self._replication_slots[name]
        else:
            self._schedule_load_slots = True
            if active:
                logger.warning("Unable to drop replication slot '%s', slot is active", name)
            else:
                logger.error("Failed to drop replication slot '%s'", name)

    def _drop_incorrect_slots(self, cluster: Cluster, slots: Dict[str, Any]) -> None:
        """Compare required slots and configured as permanent slots with those found, dropping extraneous ones.

        .. note::
            Slots that are not contained in *slots* will be dropped.
            Slots can be filtered out with ``ignore_slots`` configuration.

            Slots that have matching names but do not match attributes in *slots* will also be dropped.
            Slots that are presented in *slots* but have wal_status=lost are also dropped.

        :param cluster: cluster state information object.
        :param slots: dictionary of desired slot names as keys with slot attributes as a dictionary value, if known.
        """
        # drop old replication slots which are not presented in desired slots.
        for name, value in list(self._replication_slots.items()):
            wal_status_lost = name in slots and \
                self._postgresql.major_version >= 130000 and value['wal_status'] == 'lost'

            if name in slots and not wal_status_lost:
                continue

            if not global_config.is_paused and not self.ignore_replication_slot(cluster, name):
                if wal_status_lost:
                    logger.info("Trying to drop replication slot '%s' with wal_status=lost", name)
                else:
                    logger.info("Trying to drop unknown replication slot '%s'", name)
                self._drop_replication_slot(name)

        # drop slots with matching names but attributes that do not match, e.g. `plugin` or `database`.
        for name, value in slots.items():
            if name in self._replication_slots and not compare_slots(value, self._replication_slots[name]):
                logger.info("Trying to drop replication slot '%s' because value is changing from %s to %s",
                            name, self._replication_slots[name], value)
                if self.drop_replication_slot(name) == (False, True):
                    self._replication_slots.pop(name)
                else:
                    logger.error("Failed to drop replication slot '%s'", name)
                    self._schedule_load_slots = True

    def _ensure_physical_slots(self, slots: Dict[str, Any], clean_inactive_physical_slots: bool) -> None:
        """Create or advance physical replication *slots*.

        Any failures are logged and do not interrupt creation of all *slots*.

        :param slots: A dictionary mapping slot name to slot attributes. This method only considers a slot
                      if the value is a dictionary with the key ``type`` and a value of ``physical``.
        :param clean_inactive_physical_slots: whether replication slots with ``xmin`` and not expected
                                              to be active should be dropped.
        """
        immediately_reserve = ', true' if self._postgresql.major_version >= 90600 else ''
        for name, value in slots.items():
            if value['type'] != 'physical':
                continue
            # First we want to detect physical replication slots that are not
            # expected to be active but have NOT NULL xmin value and drop them.
            # As the slot is not expected to be active, nothing would be consuming this
            # slot, consequently no hot-standby feedback messages would be received
            # by Postgres regarding this slot. In that case, the `xmin` value would never
            # change, which would prevent Postgres from advancing the xmin horizon.
            if self._postgresql.can_advance_slots and name in self._replication_slots and\
                    self._replication_slots[name]['type'] == 'physical':
                self.copy_slot_items(self._replication_slots[name], value, is_logical=False)
                if clean_inactive_physical_slots and value.get('expected_active') is False and value['xmin']:
                    logger.warning('Dropping physical replication slot %s because of its xmin value %s',
                                   name, value['xmin'])
                    self._drop_replication_slot(name)

            # Now we will create physical replication slots that are missing.
            if name not in self._replication_slots:
                try:
                    self._query(f"SELECT pg_catalog.pg_create_physical_replication_slot(%s{immediately_reserve})"
                                f" WHERE NOT EXISTS (SELECT 1 FROM pg_catalog.pg_replication_slots"
                                f" WHERE slot_type = 'physical' AND slot_name = %s)",
                                name, name)
                except Exception:
                    logger.exception("Failed to create physical replication slot '%s'", name)
                self._schedule_load_slots = True
            # And advance restart_lsn on physical replication slots that are not expected to be active.
            elif self._postgresql.can_advance_slots and self._replication_slots[name]['type'] == 'physical' and\
                    value.get('expected_active') is not True and not value['xmin']:
                restart_lsn = value.get('restart_lsn')
                if not restart_lsn:
                    # This slot either belongs to a member or was configured as a permanent slot. However, for some
                    # reason the slot was created by an external agent instead of by Patroni, and it was created without
                    # reserving the LSN. We drop the slot here, as we cannot advance it, and let Patroni recreate and
                    # manage it in the next cycle.
                    logger.warning('Dropping physical replication slot %s because it has no restart_lsn. '
                                   'This slot was probably not created by Patroni, but by an external agent.', name)
                    self._drop_replication_slot(name)
                    continue
                lsn = value.get('lsn')
                if lsn and lsn > restart_lsn:  # The slot has feedback in DCS and needs to be advanced
                    try:
                        lsn = format_lsn(lsn)
                        self._query("SELECT pg_catalog.pg_replication_slot_advance(%s, %s)", name, lsn)
                    except Exception as exc:
                        logger.error("Error while advancing replication slot %s to position '%s': %r", name, lsn, exc)

    @contextmanager
    def get_local_connection_cursor(self, **kwargs: Any) -> Generator[Union['cursor', 'Cursor[Any]'], None, None]:
        """Create a new database connection to local server.

        Create a non-blocking connection cursor to avoid the situation where an execution of the query of
        ``pg_replication_slot_advance`` takes longer than the timeout on a HA loop, which could cause a false
        failure state.

        :param kwargs: Any keyword arguments to pass to :func:`psycopg.connect`.

        :yields: connection cursor object, note implementation varies depending on version of :mod:`psycopg`.
        """
        conn_kwargs = {**self._postgresql.connection_pool.conn_kwargs, **kwargs}
        with get_connection_cursor(**conn_kwargs) as cur:
            yield cur

    def _ensure_logical_slots_primary(self, slots: Dict[str, Any]) -> None:
        """Create any missing logical replication *slots* on the primary.

        If the logical slot already exists, copy state information into the replication slots structure stored in the
        class instance.

        :param slots: Slots that should exist are supplied in a dictionary, mapping slot name to any attributes.
                      The method will only consider slots that have a value that is a dictionary with a key ``type``
                      with a value that is ``logical``.

        """
        # Group logical slots to be created by database name
        logical_slots: Dict[str, Dict[str, Dict[str, Any]]] = defaultdict(dict)
        for name, value in slots.items():
            if value['type'] == 'logical':
                if self._replication_slots.get(name, {}).get('datoid'):
                    self.copy_slot_items(self._replication_slots[name], value)
                else:
                    logical_slots[value['database']][name] = value

        # Create new logical slots
        for database, values in logical_slots.items():
            with self.get_local_connection_cursor(dbname=database) as cur:
                for name, value in values.items():
                    try:
                        cur.execute("SELECT pg_catalog.pg_create_logical_replication_slot(%s, %s)"
                                    " WHERE NOT EXISTS (SELECT 1 FROM pg_catalog.pg_replication_slots"
                                    " WHERE slot_type = 'logical' AND slot_name = %s)",
                                    (name, value['plugin'], name))
                    except Exception as e:
                        logger.error("Failed to create logical replication slot '%s' plugin='%s': %r",
                                     name, value['plugin'], e)
                        slots.pop(name)
                    self._schedule_load_slots = True

    def _ensure_logical_slots_replica(self, slots: Dict[str, Any]) -> List[str]:
        """Update logical *slots* on replicas.

        If the logical slot already exists, copy state information into the replication slots structure stored in the
        class instance. Slots that exist are also advanced if their ``confirmed_flush_lsn`` is smaller than the state
        of the slot stored in DCS or at least the ``replay_lsn`` of this replica.

        As logical slots can only be created when the primary is available, pass the list of slots that need to be
        copied back to the caller. They will be created on replicas with :meth:`SlotsHandler.copy_logical_slots`.

        :param slots: A dictionary mapping slot name to slot attributes. This method only considers a slot
                      if the value is a dictionary with the key ``type`` and a value of ``logical``.

        :returns: list of slots to be copied from the primary.
        """
        # Group logical slots to be advanced by database name
        advance_slots: Dict[str, Dict[str, int]] = defaultdict(dict)
        create_slots: List[str] = []  # Collect logical slots to be created on the replica

        for name, value in slots.items():
            if value['type'] != 'logical':
                continue

            # If the logical already exists, copy some information about it into the original structure
            if name in self._replication_slots and compare_slots(value, self._replication_slots[name]):
                self.copy_slot_items(self._replication_slots[name], value)

                # The slot has feedback in DCS
                if 'lsn' in value:
                    # we can not advance past replay_lsn
                    advance_value = min(value['lsn'], self._postgresql.replay_lsn())
                    # Skip slots that don't need to be advanced
                    if value['confirmed_flush_lsn'] < advance_value:
                        advance_slots[value['database']][name] = advance_value
            elif name not in self._replication_slots and 'lsn' in value:
                # We want to copy only slots with feedback in a DCS
                create_slots.append(name)

        # Slots to be copied from the primary should be removed from the *slots* structure,
        # otherwise Patroni falsely assumes that they already exist.
        for name in create_slots:
            slots.pop(name)

        error, copy_slots = self._advance.schedule(advance_slots)
        if error:
            self._schedule_load_slots = True
        return create_slots + copy_slots

    def sync_replication_slots(self, cluster: Cluster, tags: Tags) -> List[str]:
        """During the HA loop read, check and alter replication slots found in the cluster.

        Read physical and logical slots from ``pg_replication_slots``, then compare to those configured in the DCS.
        Drop any slots that do not match those required by configuration and are not configured as permanent.
        Create any missing physical slots, or advance their position according to feedback stored in DCS.
        If we are the primary then create logical slots, otherwise if logical slots are known and active create
        them on replica nodes by copying slot files from the primary.

        :param cluster: object containing stateful information for the cluster.
        :param tags: reference to an object implementing :class:`Tags` interface.

        :returns: list of logical replication slots names that should be copied from the primary.
        """
        ret = []
        if self._postgresql.major_version >= 90400 and cluster.config:
            try:
                self.load_replication_slots()

                slots = cluster.get_replication_slots(self._postgresql, tags, show_error=True)

                self._drop_incorrect_slots(cluster, slots)

                # We don't want to clean physical replication slots with xmin feedback if:
                # - cluster has no leader
                # - current node is a leader, but still running as a standby
                clean_inactive_physical_slots = not cluster.is_unlocked() and \
                    (cluster.leader and cluster.leader.name != self._postgresql.name or self._postgresql.is_primary())
                self._ensure_physical_slots(slots, clean_inactive_physical_slots)

                if self._postgresql.is_primary():
                    self._logical_slots_processing_queue.clear()
                    self._ensure_logical_slots_primary(slots)
                else:
                    self.check_logical_slots_readiness(cluster, tags)
                    ret = self._ensure_logical_slots_replica(slots)

                self._replication_slots = slots
            except Exception:
                logger.exception('Exception when changing replication slots')
                self._schedule_load_slots = True
        return ret

    @contextmanager
    def _get_leader_connection_cursor(self, leader: Leader) -> Generator[Union['cursor', 'Cursor[Any]'], None, None]:
        """Create a new database connection to the leader.

        .. note::
            Uses rewind user credentials because it has enough permissions to read files from PGDATA.

        :param leader: object with information on the leader

        :yields: connection cursor object, note implementation varies depending on version of ``psycopg``.
        """
        conn_kwargs = leader.conn_kwargs(self._postgresql.config.rewind_credentials)
        conn_kwargs['dbname'] = self._postgresql.database
        with get_connection_cursor(**conn_kwargs) as cur:
            yield cur

    def check_logical_slots_readiness(self, cluster: Cluster, tags: Tags) -> bool:
        """Determine whether all known logical slots are synchronised from the leader.

        1) Retrieve the current ``catalog_xmin`` value for the physical slot from the cluster leader, and
        2) using previously stored list of "unready" logical slots, those which have yet to be checked hence have no
           stored slot attributes,
        3) store logical slot ``catalog_xmin`` when the physical slot ``catalog_xmin`` becomes valid.

        :param cluster: object containing stateful information for the cluster.
        :param tags: reference to an object implementing :class:`Tags` interface.

        :returns: ``False`` if any issue while checking logical slots readiness, ``True`` otherwise.
        """
        catalog_xmin = None
        if self._logical_slots_processing_queue and cluster.leader:
            slot_name = cluster.get_slot_name_on_primary(self._postgresql.name, tags)
            try:
                with self._get_leader_connection_cursor(cluster.leader) as cur:
                    cur.execute("SELECT slot_name, catalog_xmin FROM pg_catalog.pg_get_replication_slots()"
                                " WHERE NOT pg_catalog.pg_is_in_recovery() AND slot_name = ANY(%s)",
                                ([n for n, v in self._logical_slots_processing_queue.items()
                                  if v is None] + [slot_name],))
                    slots = {row[0]: row[1] for row in cur}
                    if slot_name not in slots:
                        logger.warning('Physical slot %s does not exist on the primary', slot_name)
                        return False
                    catalog_xmin = slots.pop(slot_name)
            except Exception as e:
                logger.error("Failed to check %s physical slot on the primary: %r", slot_name, e)
                return False

            if not self._update_pending_logical_slot_primary(slots, catalog_xmin):
                return False  # since `catalog_xmin` isn't valid further checks don't make any sense

        self._ready_logical_slots(catalog_xmin)
        return True

    def _update_pending_logical_slot_primary(self, slots: Dict[str, Any], catalog_xmin: Optional[int] = None) -> bool:
        """Store pending logical slot information for ``catalog_xmin`` on the primary.

        Remember ``catalog_xmin`` of logical slots on the primary when ``catalog_xmin`` of the physical slot became
        valid. Logical slots on replica will be safe to use after promote when ``catalog_xmin`` of the physical slot
        overtakes these values.

        :param slots: dictionary of slot information from the primary
        :param catalog_xmin: ``catalog_xmin`` of the physical slot used by this replica to stream changes from primary.

        :returns: ``False`` if any issue was faced while processing, ``True`` otherwise.
        """
        if catalog_xmin is not None:
            for name, value in slots.items():
                self._logical_slots_processing_queue[name] = value
            return True

        # Replica isn't streaming or the hot_standby_feedback isn't enabled
        try:
            if not self._query("SELECT pg_catalog.current_setting('hot_standby_feedback')::boolean")[0][0]:
                logger.error('Logical slot failover requires "hot_standby_feedback". Please check postgresql.auto.conf')
        except Exception as e:
            logger.error('Failed to check the hot_standby_feedback setting: %r', e)
        return False

    def _ready_logical_slots(self, primary_physical_catalog_xmin: Optional[int] = None) -> None:
        """Ready logical slots by comparing primary physical slot ``catalog_xmin`` to logical ``catalog_xmin``.

        The logical slot on a replica is safe to use when the physical replica slot on the primary:

            1. has a nonzero/non-null ``catalog_xmin`` represented by ``primary_physical_xmin``.
            2. has a ``catalog_xmin`` that is not newer (greater) than the ``catalog_xmin`` of any slot on the standby
            3. overtook the ``catalog_xmin`` of remembered values of logical slots on the primary.

        :param primary_physical_catalog_xmin: is the value retrieved from ``pg_catalog.pg_get_replication_slots()`` for
                                              the physical replication slot on the primary.
        """
        # Make a copy of processing queue keys as a list as the queue dictionary is modified inside the loop.
        for name in list(self._logical_slots_processing_queue):
            primary_logical_catalog_xmin = self._logical_slots_processing_queue[name]
            standby_logical_slot = self._replication_slots.get(name, {})
            standby_logical_catalog_xmin = standby_logical_slot.get('catalog_xmin', 0)
            if TYPE_CHECKING:  # pragma: no cover
                assert primary_logical_catalog_xmin is not None

            if (
                    not standby_logical_slot
                    or primary_physical_catalog_xmin is not None
                    and primary_logical_catalog_xmin <= primary_physical_catalog_xmin <= standby_logical_catalog_xmin
            ):

                del self._logical_slots_processing_queue[name]

                if standby_logical_slot:
                    logger.info('Logical slot %s is safe to be used after a failover', name)

    def copy_logical_slots(self, cluster: Cluster, tags: Tags, create_slots: List[str]) -> None:
        """Create logical replication slots on standby nodes.

        :param cluster: object containing stateful information for the cluster.
        :param tags: reference to an object implementing :class:`Tags` interface.
        :param create_slots: list of slot names to copy from the primary.
        """
        leader = cluster.leader
        if not leader:
            return
        slots = cluster.get_replication_slots(self._postgresql, tags, role=PostgresqlRole.REPLICA)
        copy_slots: Dict[str, Dict[str, Any]] = {}
        with self._get_leader_connection_cursor(leader) as cur:
            try:
                filter_failover = ' NOT failover AND' if self._postgresql.major_version >= 170000 else ''
                cur.execute("SELECT slot_name, slot_type, datname, plugin, catalog_xmin, "
                            "pg_catalog.pg_wal_lsn_diff(confirmed_flush_lsn, '0/0')::bigint, "
                            "pg_catalog.pg_read_binary_file('pg_replslot/' || slot_name || '/state')"
                            " FROM pg_catalog.pg_get_replication_slots() JOIN pg_catalog.pg_database ON datoid = oid"
                            f" WHERE{filter_failover} NOT pg_catalog.pg_is_in_recovery()"
                            " AND slot_name = ANY(%s)", (create_slots,))

                for r in cur:
                    if r[0] in slots:  # slot_name is defined in the global configuration
                        slot = {'type': r[1], 'database': r[2], 'plugin': r[3],
                                'catalog_xmin': r[4], 'confirmed_flush_lsn': r[5], 'data': r[6]}
                        if compare_slots(slot, slots[r[0]]):
                            copy_slots[r[0]] = slot
                        else:
                            logger.warning('Will not copy the logical slot "%s" due to the configuration mismatch: '
                                           'configuration=%s, slot on the primary=%s', r[0], slots[r[0]], slot)
            except Exception as e:
                logger.error("Failed to copy logical slots from the %s via postgresql connection: %r", leader.name, e)

        if copy_slots and self._postgresql.stop():
            if self._advance:
                self._advance.clean()
            pg_perm.set_permissions_from_data_directory(self._postgresql.data_dir)
            for name, value in copy_slots.items():
                slot_dir = os.path.join(self.pg_replslot_dir, name)
                slot_tmp_dir = slot_dir + '.tmp'
                if os.path.exists(slot_tmp_dir):
                    shutil.rmtree(slot_tmp_dir)
                os.makedirs(slot_tmp_dir)
                os.chmod(slot_tmp_dir, pg_perm.dir_create_mode)
                fsync_dir(slot_tmp_dir)
                slot_filename = os.path.join(slot_tmp_dir, 'state')
                with open(slot_filename, 'wb') as f:
                    os.chmod(slot_filename, pg_perm.file_create_mode)
                    f.write(value['data'])
                    f.flush()
                    os.fsync(f.fileno())
                if os.path.exists(slot_dir):
                    shutil.rmtree(slot_dir)
                os.rename(slot_tmp_dir, slot_dir)
                os.chmod(slot_dir, pg_perm.dir_create_mode)
                fsync_dir(slot_dir)
                self._logical_slots_processing_queue[name] = None
            fsync_dir(self.pg_replslot_dir)
            self._postgresql.start()

    def schedule(self, value: Optional[bool] = None) -> None:
        """Schedule the loading of slot information from the database.

        :param value: the optional value can be used to unschedule if set to ``False`` or force it to be ``True``.
                      If it is omitted the value will be ``True`` if this PostgreSQL node supports slot replication.
        """
        if value is None:
            value = self._postgresql.major_version >= 90400
        self._schedule_load_slots = self._force_readiness_check = value

    def on_promote(self) -> None:
        """Entry point from HA cycle used when a standby node is to be promoted to primary.

        .. note::
            If logical replication slot synchronisation is enabled then slot advancement will be triggered.
            If any logical slots that were copied are yet to be confirmed as ready a warning message will be logged.

        """
        if self._advance:
            self._advance.clean()

        if self._logical_slots_processing_queue:
            logger.warning('Logical replication slots that might be unsafe to use after promote: %s',
                           set(self._logical_slots_processing_queue))



import base64
import functools
import json
import logging
import os
import socket
import sys
import time

from collections import defaultdict
from enum import IntEnum
from threading import Condition, Lock, Thread
from typing import Any, Callable, Collection, Dict, Iterator, List, Optional, Tuple, Type, TYPE_CHECKING, Union

import etcd
import urllib3

from urllib3.exceptions import ProtocolError, ReadTimeoutError

from ..collections import EMPTY_DICT
from ..exceptions import DCSError, PatroniException
from ..postgresql.mpp import AbstractMPP
from ..utils import deep_compare, enable_keepalive, iter_response_objects, \
    parse_bool, RetryFailedError, USER_AGENT, WHITESPACE_RE
from . import catch_return_false_exception, Cluster, ClusterConfig, \
    Failover, Leader, Member, Status, SyncState, TimelineHistory
from .etcd import AbstractEtcd, AbstractEtcdClientWithFailover, catch_etcd_errors, \
    DnsCachingResolver, Retry, StaleEtcdNode, StaleEtcdNodeGuard

logger = logging.getLogger(__name__)


class Etcd3Error(DCSError):
    pass


class UnsupportedEtcdVersion(PatroniException):
    pass


# google.golang.org/grpc/codes
class GRPCCode(IntEnum):
    OK = 0
    Canceled = 1
    Unknown = 2
    InvalidArgument = 3
    DeadlineExceeded = 4
    NotFound = 5
    AlreadyExists = 6
    PermissionDenied = 7
    ResourceExhausted = 8
    FailedPrecondition = 9
    Aborted = 10
    OutOfRange = 11
    Unimplemented = 12
    Internal = 13
    Unavailable = 14
    DataLoss = 15
    Unauthenticated = 16


GRPCcodeToText: Dict[int, str] = {v: k for k, v in GRPCCode.__dict__['_member_map_'].items()}


class Etcd3Exception(etcd.EtcdException):
    pass


class Etcd3WatchCanceled(Etcd3Exception):
    pass


class Etcd3ClientError(Etcd3Exception):

    def __init__(self, code: Optional[int] = None, error: Optional[str] = None, status: Optional[int] = None) -> None:
        if not hasattr(self, 'code'):
            self.code = code
        if not hasattr(self, 'error'):
            self.error = error and error.strip()
        self.codeText = GRPCcodeToText.get(code) if code is not None else None
        self.status = status

    def __repr__(self) -> str:
        return "<{0} error: '{1}', code: {2}>"\
            .format(self.codeText, getattr(self, 'error', None), getattr(self, 'code', None))

    __str__ = __repr__

    def as_dict(self) -> Dict[str, Any]:
        return {'error': getattr(self, 'error', None), 'code': getattr(self, 'code', None),
                'codeText': self.codeText, 'status': self.status}

    @classmethod
    def get_subclasses(cls) -> Iterator[Type['Etcd3ClientError']]:
        for subclass in cls.__subclasses__():
            for subsubclass in subclass.get_subclasses():
                yield subsubclass
            yield subclass


class Unknown(Etcd3ClientError):
    code = GRPCCode.Unknown


class InvalidArgument(Etcd3ClientError):
    code = GRPCCode.InvalidArgument


class DeadlineExceeded(Etcd3ClientError):
    code = GRPCCode.DeadlineExceeded
    error = "context deadline exceeded"


class NotFound(Etcd3ClientError):
    code = GRPCCode.NotFound


class FailedPrecondition(Etcd3ClientError):
    code = GRPCCode.FailedPrecondition


class Unavailable(Etcd3ClientError):
    code = GRPCCode.Unavailable


# https://github.com/etcd-io/etcd/commits/main/api/v3rpc/rpctypes/error.go
class LeaseNotFound(NotFound):
    error = "etcdserver: requested lease not found"


class UserEmpty(InvalidArgument):
    error = "etcdserver: user name is empty"


class AuthFailed(InvalidArgument):
    error = "etcdserver: authentication failed, invalid user ID or password"


class AuthOldRevision(InvalidArgument):
    error = "etcdserver: revision of auth store is old"


class PermissionDenied(Etcd3ClientError):
    code = GRPCCode.PermissionDenied
    error = "etcdserver: permission denied"


class AuthNotEnabled(FailedPrecondition):
    error = "etcdserver: authentication is not enabled"


class InvalidAuthToken(Etcd3ClientError):
    code = GRPCCode.Unauthenticated
    error = "etcdserver: invalid auth token"


errStringToClientError = {getattr(s, 'error'): s for s in Etcd3ClientError.get_subclasses() if hasattr(s, 'error')}
errCodeToClientError = {getattr(s, 'code'): s for s in Etcd3ClientError.__subclasses__()}


def _raise_for_data(data: Union[bytes, str, Dict[str, Any]], status_code: Optional[int] = None) -> Etcd3ClientError:
    try:
        if TYPE_CHECKING:  # pragma: no cover
            assert isinstance(data, dict)
        data_error: Optional[Dict[str, Any]] = data.get('error') or data.get('Error')
        if isinstance(data_error, dict):  # streaming response
            status_code = data_error.get('http_code')
            code: Optional[int] = data_error.get('code') or data_error['grpc_code']
            error: str = data_error['message']
        else:
            data_code = data.get('code') or data.get('Code')
            if TYPE_CHECKING:  # pragma: no cover
                assert not isinstance(data_code, dict)
            code = data_code
            error = str(data_error or data.get('message') or data)
    except Exception:
        error = str(data)
        code = GRPCCode.Unknown

    # Workaround for etcd-io/etcd#21671
    if code == GRPCCode.Unknown and error == 'not a primary lessor':
        code = GRPCCode.Unavailable

    err = errStringToClientError.get(error) or errCodeToClientError.get(code) or Etcd3ClientError
    return err(code, error, status_code)


def to_bytes(v: Union[str, bytes]) -> bytes:
    return v if isinstance(v, bytes) else v.encode('utf-8')


def prefix_range_end(v: str) -> bytes:
    ret = bytearray(to_bytes(v))
    for i in range(len(ret) - 1, -1, -1):
        if ret[i] < 0xff:
            ret[i] += 1
            break
    return bytes(ret)


def base64_encode(v: Union[str, bytes]) -> str:
    return base64.b64encode(to_bytes(v)).decode('utf-8')


def base64_decode(v: str) -> str:
    return base64.b64decode(v).decode('utf-8')


def build_range_request(key: str, range_end: Union[bytes, str, None] = None) -> Dict[str, Any]:
    fields = {'key': base64_encode(key)}
    if range_end:
        fields['range_end'] = base64_encode(range_end)
    return fields


def _handle_auth_errors(func: Callable[..., Any]) -> Any:
    def wrapper(self: 'Etcd3Client', *args: Any, **kwargs: Any) -> Any:
        return self.handle_auth_errors(func, *args, **kwargs)
    return wrapper


class Etcd3Client(AbstractEtcdClientWithFailover):

    ERROR_CLS = Etcd3Error

    def __init__(self, config: Dict[str, Any], dns_resolver: DnsCachingResolver, cache_ttl: int = 300) -> None:
        self._decoder = json.JSONDecoder()
        self._reauthenticate = False
        self._token = None
        self._cluster_version: Tuple[int, ...] = tuple()
        super(Etcd3Client, self).__init__({**config, 'version_prefix': '/v3beta'}, dns_resolver, cache_ttl)
        if self._use_proxies and not self._cluster_version:
            kwargs = self._prepare_common_parameters(1)
            self._ensure_version_prefix(self._base_uri, **kwargs)
            self.authenticate_on_start()

    def authenticate_on_start(self, auth_request_func: Optional[Callable[..., Dict[str, Any]]] = None):
        """Authenticate with Etcd v3 at startup and exit on invalid credentials.

        :param auth_request_func: optional custom authentication request function,
                                  if not provided, :meth:`call_rpc` will be used.
        """
        try:
            self.authenticate(auth_request_func=auth_request_func)
        except AuthFailed as e:
            logger.fatal('Etcd3 authentication failed: %r', e)
            sys.exit(1)

    def _get_headers(self) -> Dict[str, str]:
        headers = urllib3.make_headers(user_agent=USER_AGENT)
        if self._token and self._cluster_version >= (3, 3, 0):
            headers['authorization'] = self._token
        return headers

    def _prepare_request(self, kwargs: Dict[str, Any], params: Optional[Dict[str, Any]] = None,
                         method: Optional[str] = None) -> Callable[..., urllib3.response.HTTPResponse]:
        if params is not None:
            kwargs['body'] = json.dumps(params)
            kwargs['headers']['Content-Type'] = 'application/json'
        return self.http.urlopen

    def _handle_server_response(self, response: urllib3.response.HTTPResponse) -> Dict[str, Any]:
        data = response.data
        try:
            data = data.decode('utf-8')
            idx = WHITESPACE_RE.match(data, 0).end()  # pyright: ignore [reportOptionalMemberAccess]
            ret: Dict[str, Any] = self._decoder.raw_decode(data, idx)[0]

            header = ret.get('header', EMPTY_DICT)
            self._check_cluster_raft_term(header.get('cluster_id'), header.get('raft_term'))

            if response.status < 400:
                return ret
        except (TypeError, ValueError, UnicodeError) as e:
            if response.status < 400:
                raise etcd.EtcdException("Server response '%s' was not valid JSON: %r" % (data, e))
            ret = {}
        ex = _raise_for_data(ret or data, response.status)
        if isinstance(ex, Unavailable):
            # <Unavailable error: 'etcdserver: request timed out', code: 14>
            # Pretend that we got `socket.timeout` and let `AbstractEtcdClientWithFailover._do_http_request()`
            # method handle it by switching to another etcd node and retrying request.
            raise socket.timeout from ex
        else:
            raise ex

    def _ensure_version_prefix(self, base_uri: str, **kwargs: Any) -> None:
        if self.version_prefix != '/v3':
            response = self.http.urlopen(self._MGET, base_uri + '/version', **kwargs)
            response = self._handle_server_response(response)

            server_version_str = response['etcdserver']
            server_version = tuple(int(x) for x in server_version_str.split('.'))
            cluster_version_str = response['etcdcluster']
            self._cluster_version = tuple(int(x) for x in cluster_version_str.split('.'))

            if self._cluster_version < (3, 0) or server_version < (3, 0, 4):
                raise UnsupportedEtcdVersion('Detected Etcd version {0} is lower than 3.0.4'.format(server_version_str))

            if self._cluster_version < (3, 3):
                if self.version_prefix != '/v3alpha':
                    if self._cluster_version < (3, 1):
                        logger.warning('Detected Etcd version %s is lower than 3.1.0, watches are not supported',
                                       cluster_version_str)
                    if self.username and self.password:
                        logger.warning('Detected Etcd version %s is lower than 3.3.0, authentication is not supported',
                                       cluster_version_str)
                    self.version_prefix = '/v3alpha'
            elif self._cluster_version < (3, 4):
                self.version_prefix = '/v3beta'
            else:
                self.version_prefix = '/v3'

    def _prepare_get_members(self, etcd_nodes: int) -> Dict[str, Any]:
        kwargs = self._prepare_common_parameters(etcd_nodes)
        self._prepare_request(kwargs, {})
        return kwargs

    def _do_auth_request(self, base_uri: str, kwargs: Dict[str, Any],
                         method: str, fields: Dict[str, Any], retry: Optional[Retry] = None) -> Dict[str, Any]:
        """Special method for handling authentication when discovering cluster members.

        We can't use `call_rpc()` method for this purpose because it may cause infinite recursion.

        :param base_uri: base url for authentication request, e.g. `http://etcd:2379/v3`
        :param kwargs: common request parameters, e.g. headers.
        :param method: `/auth/authenticate`
        :param fields: authentication fields, e.g. `{'name': 'user', 'password': 'pass'}`.
        :param retry: optional retry configuration, ignored.
        """
        request_kwargs = kwargs.copy()
        request_kwargs['headers'] = {k: v for k, v in kwargs['headers'].items() if k != 'authorization'}
        self._prepare_request(request_kwargs, fields)
        response = self.http.urlopen(self._MPOST, base_uri + method, **request_kwargs)
        return self._handle_server_response(response)

    def _do_member_list_request(self, url: str, retry: Optional[Retry] = None, **kwargs: Any) -> Any:
        """Special method for handling member list requests.

        :param url: base url for member list request, e.g. `http://etcd:2379/v3/cluster/member/list`
        :param kwargs: common request parameters, e.g. headers.
        :param retry: optional retry configuration, ignored.
        """
        request_kwargs = kwargs.copy()
        request_kwargs['headers'] = kwargs['headers'].copy()
        # We want to update headers with authentication token if it was obtained during authentication request.
        request_kwargs['headers'].update(self._get_headers())
        response = self.http.urlopen(self._MPOST, url, **request_kwargs)
        return self._handle_server_response(response)

    def _get_members(self, base_uri: str, **kwargs: Any) -> List[str]:
        self._ensure_version_prefix(base_uri, **kwargs)
        base_uri += self.version_prefix

        retry = None
        if self._update_machines_cache:
            retry = Retry(deadline=self._config['retry_timeout'], max_delay=1, max_tries=-1)
            # handle_auth_errors() calls retry.ensure_deadline(), which expects Retry.__call__()
            # to have initialized the internal deadline state first.
            retry(lambda: None)

        # custom authentication request function, because we can't use `call_rpc()` method
        # for this purpose as it may cause infinite recursion
        auth_request_func = functools.partial(self._do_auth_request, base_uri, kwargs)

        # if _machine_cache is empty, it means we are just starting and want to exit early if authentication fails.
        if not self._machines_cache:
            self.authenticate_on_start(auth_request_func)

        response = self.handle_auth_errors(Etcd3Client._do_member_list_request, base_uri + '/cluster/member/list',
                                           auth_request_func=auth_request_func, retry=retry, **kwargs)
        return [url for member in response['members'] for url in member.get('clientURLs', [])]

    def call_rpc(self, method: str, fields: Dict[str, Any], retry: Optional[Retry] = None) -> Dict[str, Any]:
        fields['retry'] = retry
        return self.api_execute(self.version_prefix + method, self._MPOST, fields)

    def authenticate(self, *, retry: Optional[Retry] = None,
                     auth_request_func: Optional[Callable[..., Dict[str, Any]]] = None) -> bool:
        """Authenticate with the Etcd v3 cluster.

        :param retry: optional retry configuration.
        :param auth_request_func: optional custom authentication request function,
                                  if not provided, `call_rpc()` method will be used.
        """
        if not (self._cluster_version >= (3, 3) and self.username and self.password):
            return False
        if not auth_request_func:
            auth_request_func = self.call_rpc
        logger.info('Trying to authenticate on Etcd...')
        old_token, self._token = self._token, None
        try:
            response = auth_request_func('/auth/authenticate',
                                         {'name': self.username, 'password': self.password}, retry)
        except AuthNotEnabled:
            logger.info('Etcd authentication is not enabled')
            self._token = None
        except Exception:
            self._token = old_token
            raise
        else:
            self._token = response.get('token')
        return old_token != self._token

    def handle_auth_errors(self: 'Etcd3Client', func: Callable[..., Any], *args: Any,
                           auth_request_func: Optional[Callable[..., Dict[str, Any]]] = None,
                           retry: Optional[Retry] = None, **kwargs: Any) -> Any:
        """Handle authentication errors for the given function.

        :param func: function to call.
        :param args: positional arguments for the function.
        :param auth_request_func: optional custom authentication request function,
                                  if not provided, `call_rpc()` method will be used.
        :param retry: optional retry configuration.
        :param kwargs: keyword arguments for the function.
        """
        reauthenticated = False
        exc = None
        while True:
            if self._reauthenticate:
                if self.username and self.password:
                    self.authenticate(retry=retry, auth_request_func=auth_request_func)
                    self._reauthenticate = False
                else:
                    msg = 'Username or password not set, authentication is not possible'
                    logger.fatal(msg)
                    raise exc or Etcd3Exception(msg)
                reauthenticated = True

            try:
                return func(self, *args, retry=retry, **kwargs)
            except (UserEmpty, PermissionDenied) as e:  # no token provided
                # PermissionDenied is raised on 3.0 and 3.1
                if self._cluster_version < (3, 3) and (not isinstance(e, PermissionDenied)
                                                       or self._cluster_version < (3, 2)):
                    raise UnsupportedEtcdVersion('Authentication is required by Etcd cluster but not '
                                                 'supported on version lower than 3.3.0. Cluster version: '
                                                 '{0}'.format('.'.join(map(str, self._cluster_version))))
                exc = e
            except InvalidAuthToken as e:
                logger.error('Invalid auth token: %s', self._token)
                exc = e
            except AuthOldRevision as e:
                logger.error('Auth token is for old revision of auth store')
                exc = e
            self._reauthenticate = True
            if retry:
                retry.ensure_deadline(0.5, exc)
            elif reauthenticated:
                raise exc

    @_handle_auth_errors
    def range(self, key: str, range_end: Union[bytes, str, None] = None, serializable: bool = True,
              *, retry: Optional[Retry] = None) -> Dict[str, Any]:
        params = build_range_request(key, range_end)
        params['serializable'] = serializable  # For better performance. We can tolerate stale reads
        return self.call_rpc('/kv/range', params, retry)

    def prefix(self, key: str, serializable: bool = True, *, retry: Optional[Retry] = None) -> Dict[str, Any]:
        return self.range(key, prefix_range_end(key), serializable, retry=retry)

    @_handle_auth_errors
    def lease_grant(self, ttl: int, *, retry: Optional[Retry] = None) -> str:
        return self.call_rpc('/lease/grant', {'TTL': ttl}, retry)['ID']

    @_handle_auth_errors
    def lease_keepalive(self, ID: str, *, retry: Optional[Retry] = None) -> Optional[str]:
        return self.call_rpc('/lease/keepalive', {'ID': ID}, retry).get('result', {}).get('TTL')

    @_handle_auth_errors
    def txn(self, compare: Dict[str, Any], success: Dict[str, Any],
            failure: Optional[Dict[str, Any]] = None, *, retry: Optional[Retry] = None) -> Dict[str, Any]:
        fields = {'compare': [compare], 'success': [success]}
        if failure:
            fields['failure'] = [failure]
        ret = self.call_rpc('/kv/txn', fields, retry)
        return ret if failure or ret.get('succeeded') else {}

    @_handle_auth_errors
    def put(self, key: str, value: str, lease: Optional[str] = None, create_revision: Optional[str] = None,
            mod_revision: Optional[str] = None, *, retry: Optional[Retry] = None) -> Dict[str, Any]:
        fields = {'key': base64_encode(key), 'value': base64_encode(value)}
        if lease:
            fields['lease'] = lease
        if create_revision is not None:
            compare = {'target': 'CREATE', 'create_revision': create_revision}
        elif mod_revision is not None:
            compare = {'target': 'MOD', 'mod_revision': mod_revision}
        else:
            return self.call_rpc('/kv/put', fields, retry)
        compare['key'] = fields['key']
        return self.txn(compare, {'request_put': fields}, retry=retry)

    @_handle_auth_errors
    def deleterange(self, key: str, range_end: Union[bytes, str, None] = None,
                    mod_revision: Optional[str] = None, *, retry: Optional[Retry] = None) -> Dict[str, Any]:
        fields = build_range_request(key, range_end)
        if mod_revision is None:
            return self.call_rpc('/kv/deleterange', fields, retry)
        compare = {'target': 'MOD', 'mod_revision': mod_revision, 'key': fields['key']}
        return self.txn(compare, {'request_delete_range': fields}, retry=retry)

    def deleteprefix(self, key: str, *, retry: Optional[Retry] = None) -> Dict[str, Any]:
        return self.deleterange(key, prefix_range_end(key), retry=retry)

    def watchrange(self, key: str, range_end: Union[bytes, str, None] = None,
                   start_revision: Optional[str] = None, filters: Optional[List[Dict[str, Any]]] = None,
                   read_timeout: Optional[float] = None) -> urllib3.response.HTTPResponse:
        """returns: response object"""
        params = build_range_request(key, range_end)
        if start_revision is not None:
            params['start_revision'] = start_revision
        params['filters'] = filters or []
        kwargs = self._prepare_common_parameters(1, self.read_timeout)
        request_executor = self._prepare_request(kwargs, {'create_request': params})
        kwargs.update(timeout=urllib3.Timeout(connect=kwargs['timeout'], read=read_timeout), retries=0)
        return request_executor(self._MPOST, self._base_uri + self.version_prefix + '/watch', **kwargs)

    def watchprefix(self, key: str, start_revision: Optional[str] = None,
                    filters: Optional[List[Dict[str, Any]]] = None,
                    read_timeout: Optional[float] = None) -> urllib3.response.HTTPResponse:
        return self.watchrange(key, prefix_range_end(key), start_revision, filters, read_timeout)


class KVCache(StaleEtcdNodeGuard, Thread):

    def __init__(self, dcs: 'Etcd3', client: 'PatroniEtcd3Client') -> None:
        Thread.__init__(self)
        StaleEtcdNodeGuard.__init__(self)
        self.daemon = True
        self._dcs = dcs
        self._client = client
        self.condition = Condition()
        self._config_key = base64_encode(dcs.config_path)
        self._leader_key = base64_encode(dcs.leader_path)
        self._optime_key = base64_encode(dcs.leader_optime_path)
        self._status_key = base64_encode(dcs.status_path)
        self._name = base64_encode(getattr(dcs, '_name'))  # pyright
        self._is_ready = False
        self._response = None
        self._response_lock = Lock()
        self._object_cache = {}
        self._object_cache_lock = Lock()
        self.start()

    def set(self, value: Dict[str, Any], overwrite: bool = False) -> Tuple[bool, Optional[Dict[str, Any]]]:
        with self._object_cache_lock:
            name = value['key']
            old_value = self._object_cache.get(name)
            ret = not old_value or int(old_value['mod_revision']) < int(value['mod_revision'])
            if ret or overwrite and old_value and old_value['mod_revision'] == value['mod_revision']:
                self._object_cache[name] = value
        return ret, old_value

    def delete(self, name: str, mod_revision: str) -> Tuple[bool, Optional[Dict[str, Any]]]:
        with self._object_cache_lock:
            old_value = self._object_cache.get(name)
            ret = old_value and int(old_value['mod_revision']) < int(mod_revision)
            if ret:
                del self._object_cache[name]
        return bool(not old_value or ret), old_value

    def copy(self) -> List[Dict[str, Any]]:
        with self._object_cache_lock:
            return [v.copy() for v in self._object_cache.values()]

    def get(self, name: str) -> Optional[Dict[str, Any]]:
        with self._object_cache_lock:
            return self._object_cache.get(name)

    def _process_event(self, event: Dict[str, Any]) -> None:
        kv = event['kv']
        key = kv['key']
        if event.get('type') == 'DELETE':
            success, old_value = self.delete(key, kv['mod_revision'])
        else:
            success, old_value = self.set(kv, True)

        if success:
            old_value = old_value and old_value.get('value')
            new_value = kv.get('value')

            value_changed = old_value != new_value and \
                (key == self._leader_key or key in (self._optime_key, self._status_key) and new_value is not None
                 or key == self._config_key and old_value is not None and new_value is not None)

            if value_changed:
                logger.debug('%s changed from %s to %s', key, old_value, new_value)

            # We also want to wake up HA loop on replicas if leader optime (or status key) was updated
            if value_changed and (key not in (self._optime_key, self._status_key)
                                  or (self.get(self._leader_key) or {}).get('value') != self._name):
                self._dcs.event.set()

    def _process_message(self, message: Dict[str, Any]) -> None:
        logger.debug('Received message: %s', message)
        if 'error' in message:
            raise _raise_for_data(message)
        result = message.get('result', EMPTY_DICT)
        if parse_bool(result.get('canceled')):
            raise Etcd3WatchCanceled('Watch canceled')
        header = result.get('header', EMPTY_DICT)
        self._check_cluster_raft_term(header.get('cluster_id'), header.get('raft_term'))
        events: List[Dict[str, Any]] = result.get('events', [])
        for event in events:
            self._process_event(event)

    @staticmethod
    def _finish_response(response: urllib3.response.HTTPResponse) -> None:
        try:
            response.close()
        finally:
            response.release_conn()

    def _do_watch(self, revision: str) -> None:
        with self._response_lock:
            self._response = None
        # We do most of requests with timeouts. The only exception /watch requests to Etcd v3.
        # In order to interrupt the /watch request we do socket.shutdown() from the main thread,
        # which doesn't work on Windows. Therefore we want to use the last resort, `read_timeout`.
        # Setting it to TTL will help to partially mitigate the problem.
        # Setting it to lower value is not nice because for idling clusters it will increase
        # the numbers of interrupts and reconnects.
        read_timeout = self._dcs.ttl if os.name == 'nt' else None
        response = self._client.watchprefix(self._dcs.cluster_prefix, revision, read_timeout=read_timeout)
        with self._response_lock:
            if self._response is None:
                self._response = response

        if not self._response:
            return self._finish_response(response)

        for message in iter_response_objects(response):
            self._process_message(message)

    def _build_cache(self) -> None:
        result = self._dcs.retry(self._client.prefix, self._dcs.cluster_prefix)
        header = result.get('header', EMPTY_DICT)
        with self._object_cache_lock:
            self._reset_cluster_raft_term()
            self._object_cache = {node['key']: node for node in result.get('kvs', [])}
            self._check_cluster_raft_term(header.get('cluster_id'), header.get('raft_term'))
        with self.condition:
            self._is_ready = True
            self.condition.notify()

        try:
            self._do_watch(result['header']['revision'])
        except Etcd3WatchCanceled:
            logger.info('Watch request canceled')
        except Exception as e:
            # Following exceptions are expected on Windows because the /watch request is done with `read_timeout`
            if not (os.name == 'nt' and isinstance(e, (ReadTimeoutError, ProtocolError))):
                logger.error('watchprefix failed: %r', e)
        finally:
            with self.condition:
                self._is_ready = False
            with self._response_lock:
                response, self._response = self._response, None
            if isinstance(response, urllib3.response.HTTPResponse):
                self._finish_response(response)

    def run(self) -> None:
        while True:
            try:
                self._build_cache()
            except Exception as e:
                logger.error('KVCache.run %r', e)
                time.sleep(1)

    def kill_stream(self) -> None:
        conn_sock: Any = None
        with self._response_lock:
            if isinstance(self._response, urllib3.response.HTTPResponse):
                try:
                    conn_sock = self._response.connection.sock if self._response.connection else None
                except Exception:
                    conn_sock = None
            else:
                self._response = False
        if conn_sock:
            # python-etcd forces usage of pyopenssl if the last one is available.
            # In this case HTTPConnection.socket is not inherited from socket.socket, but urllib3 uses custom
            # class `WrappedSocket`, which shutdown() method could be incompatible with socket.shutdown().
            # Therefore we use WrappedSocket.socket, which points to original `socket` object.
            sock: socket.socket = conn_sock.socket if conn_sock.__class__.__name__ == 'WrappedSocket' else conn_sock
            try:
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
            except Exception as e:
                logger.debug('Error on socket.shutdown: %r', e)

    def is_ready(self) -> bool:
        """Must be called only when holding the lock on `condition`"""
        if self._is_ready:
            try:
                self._client._check_cluster_raft_term(self._cluster_id, self._raft_term)
            except StaleEtcdNode:
                self._is_ready = False
                self.kill_stream()
        return self._is_ready


class PatroniEtcd3Client(Etcd3Client):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._kv_cache = None
        super(PatroniEtcd3Client, self).__init__(*args, **kwargs)

    def configure(self, etcd3: 'Etcd3') -> None:
        self._etcd3 = etcd3

    def start_watcher(self) -> None:
        if self._cluster_version >= (3, 1):
            self._kv_cache = KVCache(self._etcd3, self)

    def _restart_watcher(self) -> None:
        if self._kv_cache:
            self._kv_cache.kill_stream()

    def set_base_uri(self, value: str) -> None:
        super(PatroniEtcd3Client, self).set_base_uri(value)
        self._restart_watcher()

    def _wait_cache(self, timeout: float) -> None:
        stop_time = time.time() + timeout
        while self._kv_cache and not self._kv_cache.is_ready():
            timeout = stop_time - time.time()
            if timeout <= 0:
                raise RetryFailedError('Exceeded retry deadline')
            self._kv_cache.condition.wait(timeout)

    def get_cluster(self, path: str) -> List[Dict[str, Any]]:
        if self._kv_cache and path.startswith(self._etcd3.cluster_prefix):
            with self._kv_cache.condition:
                self._wait_cache(self.read_timeout)
                ret = self._kv_cache.copy()
        else:
            serializable = not getattr(self._etcd3, '_ctl')  # use linearizable for patronictl
            ret = self._etcd3.retry(self.prefix, path, serializable).get('kvs', [])
        for node in ret:
            node.update({'key': base64_decode(node['key']),
                         'value': base64_decode(node.get('value', '')),
                         'lease': node.get('lease')})
        return ret

    def call_rpc(self, method: str, fields: Dict[str, Any], retry: Optional[Retry] = None) -> Dict[str, Any]:
        ret = super(PatroniEtcd3Client, self).call_rpc(method, fields, retry)

        if self._kv_cache:
            value = delete = None
            # For the 'failure' case we only support a second (nested) transaction that attempts to
            # update/delete the same keys. Anything more complex than that we don't need and therefore it doesn't
            # make sense to write a universal response analyzer and we can just check expected JSON path.
            if method == '/kv/txn'\
                    and (ret.get('succeeded') or 'failure' in fields and 'request_txn' in fields['failure'][0]
                         and ret.get('responses', [{'response_txn': {'succeeded': False}}])[0]
                         .get('response_txn', {}).get('succeeded')):
                on_success = fields['success'][0]
                value = on_success.get('request_put')
                delete = on_success.get('request_delete_range')
            elif method == '/kv/put' and ret:
                value = fields
            elif method == '/kv/deleterange' and ret:
                delete = fields

            if value:
                value['mod_revision'] = ret['header']['revision']
                self._kv_cache.set(value)
            elif delete and 'range_end' not in delete:
                self._kv_cache.delete(delete['key'], ret['header']['revision'])

        return ret

    def txn(self, compare: Dict[str, Any], success: Dict[str, Any],
            failure: Optional[Dict[str, Any]] = None, *, retry: Optional[Retry] = None) -> Dict[str, Any]:
        ret = super(PatroniEtcd3Client, self).txn(compare, success, failure, retry=retry)
        # Here we abuse the fact that the `failure` is only set in the call from update_leader().
        # In all other cases the txn() call failure may be an indicator of a stale cache,
        # and therefore we want to restart watcher.
        if not failure and not ret:
            self._restart_watcher()
        return ret


class Etcd3(AbstractEtcd):

    def __init__(self, config: Dict[str, Any], mpp: AbstractMPP) -> None:
        super(Etcd3, self).__init__(config, mpp, PatroniEtcd3Client, (DeadlineExceeded, FailedPrecondition))
        self.__do_not_watch = False
        self._lease = None
        self._last_lease_refresh = 0

        self._client.configure(self)
        if not self._ctl:
            self._client.start_watcher()
            self.create_lease()

    @property
    def _client(self) -> PatroniEtcd3Client:
        if TYPE_CHECKING:  # pragma: no cover
            assert isinstance(self._abstract_client, PatroniEtcd3Client)
        return self._abstract_client

    def set_socket_options(self, sock: socket.socket,
                           socket_options: Optional[Collection[Tuple[int, int, int]]]) -> None:
        if TYPE_CHECKING:  # pragma: no cover
            assert self._retry.deadline is not None
        enable_keepalive(sock, self.ttl, int(self.loop_wait + self._retry.deadline))

    def set_ttl(self, ttl: int) -> Optional[bool]:
        self.__do_not_watch = super(Etcd3, self).set_ttl(ttl)
        if self.__do_not_watch:
            self._lease = None
        return None

    def _do_refresh_lease(self, force: bool = False, retry: Optional[Retry] = None) -> bool:
        if not force and self._lease and self._last_lease_refresh + self._loop_wait > time.time():
            return False

        if self._lease and not self._client.lease_keepalive(self._lease, retry=retry):
            self._lease = None

        ret = not self._lease
        if ret:
            self._lease = self._client.lease_grant(self._ttl, retry=retry)

        self._last_lease_refresh = time.time()
        return ret

    def refresh_lease(self) -> bool:
        try:
            return self.retry(self._do_refresh_lease)
        except (Etcd3ClientError, RetryFailedError):
            logger.exception('refresh_lease')
        raise Etcd3Error('Failed to keepalive/grant lease')

    def create_lease(self) -> None:
        while not self._lease:
            try:
                self.refresh_lease()
            except Etcd3Error:
                logger.info('waiting on etcd')
                time.sleep(5)

    @property
    def cluster_prefix(self) -> str:
        """Construct the cluster prefix for the cluster.

        :returns: path in the DCS under which we store information about this Patroni cluster.
        """
        return self._base_path + '/' if self.is_mpp_coordinator() else self.client_path('')

    @staticmethod
    def member(node: Dict[str, str]) -> Member:
        return Member.from_node(node['mod_revision'], os.path.basename(node['key']), node['lease'], node['value'])

    def _cluster_from_nodes(self, nodes: Dict[str, Any]) -> Cluster:
        # get initialize flag
        initialize = nodes.get(self._INITIALIZE)
        initialize = initialize and initialize['value']

        # get global dynamic configuration
        config = nodes.get(self._CONFIG)
        config = config and ClusterConfig.from_node(config['mod_revision'], config['value'])

        # get timeline history
        history = nodes.get(self._HISTORY)
        history = history and TimelineHistory.from_node(history['mod_revision'], history['value'])

        # get last know leader lsn and slots
        status = nodes.get(self._STATUS) or nodes.get(self._LEADER_OPTIME)
        status = Status.from_node(status and status['value'])

        # get list of members
        members = [self.member(n) for k, n in nodes.items() if k.startswith(self._MEMBERS) and k.count('/') == 1]

        # get leader
        leader = nodes.get(self._LEADER)
        if not self._ctl and leader and leader['value'] == self._name and self._lease != leader.get('lease'):
            logger.warning('I am the leader but not owner of the lease')

        if leader:
            member = Member(-1, leader['value'], None, {})
            member = ([m for m in members if m.name == leader['value']] or [member])[0]
            leader = Leader(leader['mod_revision'], leader['lease'], member)

        # failover key
        failover = nodes.get(self._FAILOVER)
        if failover:
            failover = Failover.from_node(failover['mod_revision'], failover['value'])

        # get synchronization state
        sync = nodes.get(self._SYNC)
        sync = SyncState.from_node(sync and sync['mod_revision'], sync and sync['value'])

        # get failsafe topology
        failsafe = nodes.get(self._FAILSAFE)
        try:
            failsafe = json.loads(failsafe['value']) if failsafe else None
        except Exception:
            failsafe = None

        return Cluster(initialize, config, leader, status, members, failover, sync, history, failsafe)

    def _postgresql_cluster_loader(self, path: str) -> Cluster:
        """Load and build the :class:`Cluster` object from DCS, which represents a single PostgreSQL cluster.

        :param path: the path in DCS where to load :class:`Cluster` from.

        :returns: :class:`Cluster` instance.
        """
        nodes = {node['key'][len(path):]: node
                 for node in self._client.get_cluster(path)
                 if node['key'].startswith(path)}
        return self._cluster_from_nodes(nodes)

    def _mpp_cluster_loader(self, path: str) -> Dict[int, Cluster]:
        """Load and build all PostgreSQL clusters from a single MPP cluster.

        :param path: the path in DCS where to load Cluster(s) from.

        :returns: all MPP groups as :class:`dict`, with group IDs as keys and :class:`Cluster` objects as values.
        """
        clusters: Dict[int, Dict[str, Dict[str, Any]]] = defaultdict(dict)
        path = self._base_path + '/'
        for node in self._client.get_cluster(path):
            key = node['key'][len(path):].split('/', 1)
            if len(key) == 2 and self._mpp.group_re.match(key[0]):
                clusters[int(key[0])][key[1]] = node
        return {group: self._cluster_from_nodes(nodes) for group, nodes in clusters.items()}

    def _load_cluster(
            self, path: str, loader: Callable[[str], Union[Cluster, Dict[int, Cluster]]]
    ) -> Union[Cluster, Dict[int, Cluster]]:
        cluster = None
        try:
            cluster = loader(path)
        except UnsupportedEtcdVersion:
            raise
        except Exception as e:
            self._handle_exception(e, 'get_cluster', raise_ex=Etcd3Error('Etcd is not responding properly'))
        self._has_failed = False
        if TYPE_CHECKING:  # pragma: no cover
            assert cluster is not None
        return cluster

    @catch_etcd_errors
    def touch_member(self, data: Dict[str, Any]) -> bool:
        try:
            self.refresh_lease()
        except Etcd3Error:
            return False

        cluster = self.cluster
        member = cluster and cluster.get_member(self._name, fallback_to_leader=False)

        if member and member.session == self._lease and deep_compare(data, member.data):
            return True

        value = json.dumps(data, separators=(',', ':'))
        try:
            return bool(self._client.put(self.member_path, value, self._lease))
        except LeaseNotFound:
            self._lease = None
            logger.error('Our lease disappeared from Etcd, can not "touch_member"')
        return False

    @catch_etcd_errors
    def take_leader(self) -> bool:
        return self.retry(self._client.put, self.leader_path, self._name, self._lease)

    def _do_attempt_to_acquire_leader(self, retry: Retry) -> bool:
        def _retry(*args: Any, **kwargs: Any) -> Any:
            kwargs['retry'] = retry
            return retry(*args, **kwargs)

        try:
            return _retry(self._client.put, self.leader_path, self._name, self._lease, create_revision='0')
        except LeaseNotFound:
            self._lease = None
            if not retry.ensure_deadline(0):
                logger.error('Our lease disappeared from Etcd. Deadline exceeded, giving up')
                return False

            logger.error('Our lease disappeared from Etcd. Will try to get a new one and retry attempt')

            _retry(self._do_refresh_lease)

            retry.ensure_deadline(1, Etcd3Error('_do_attempt_to_acquire_leader timeout'))
            return _retry(self._client.put, self.leader_path, self._name, self._lease, create_revision='0')

    @catch_return_false_exception
    def attempt_to_acquire_leader(self) -> bool:
        retry = self._retry.copy()

        def _retry(*args: Any, **kwargs: Any) -> Any:
            kwargs['retry'] = retry
            return retry(*args, **kwargs)

        self._run_and_handle_exceptions(self._do_refresh_lease, retry=_retry)

        retry.ensure_deadline(1, Etcd3Error('attempt_to_acquire_leader timeout'))

        ret = self._run_and_handle_exceptions(self._do_attempt_to_acquire_leader, retry, retry=None)
        if not ret:
            logger.info('Could not take out TTL lock')
        return ret

    @catch_etcd_errors
    def set_failover_value(self, value: str, version: Optional[str] = None) -> bool:
        return bool(self._client.put(self.failover_path, value, mod_revision=version))

    @catch_etcd_errors
    def set_config_value(self, value: str, version: Optional[str] = None) -> bool:
        return bool(self._client.put(self.config_path, value, mod_revision=version))

    @catch_etcd_errors
    def _write_leader_optime(self, last_lsn: str) -> bool:
        return bool(self._client.put(self.leader_optime_path, last_lsn))

    @catch_etcd_errors
    def _write_status(self, value: str) -> bool:
        return bool(self._client.put(self.status_path, value))

    @catch_etcd_errors
    def _write_failsafe(self, value: str) -> bool:
        return bool(self._client.put(self.failsafe_path, value))

    @catch_return_false_exception
    def _update_leader(self, leader: Leader) -> bool:
        retry = self._retry.copy()

        def _retry(*args: Any, **kwargs: Any) -> Any:
            kwargs['retry'] = retry
            return retry(*args, **kwargs)

        self._run_and_handle_exceptions(self._do_refresh_lease, True, retry=_retry)

        if self._lease and leader.session != self._lease:
            retry.ensure_deadline(1, Etcd3Error('update_leader timeout'))

            fields = {'key': base64_encode(self.leader_path), 'value': base64_encode(self._name), 'lease': self._lease}
            # First we try to update lease on existing leader key "hoping" that we still owning it
            compare1 = {'key': fields['key'], 'target': 'VALUE', 'value': fields['value']}
            request_put = {'request_put': fields}
            # If the first comparison failed we will try to create the new leader key in a transaction
            compare2 = {'key': fields['key'], 'target': 'CREATE', 'create_revision': '0'}
            request_txn = {'request_txn': {'compare': [compare2], 'success': [request_put]}}
            ret = self._run_and_handle_exceptions(self._client.txn, compare1, request_put, request_txn, retry=_retry)
            return ret.get('succeeded', False)\
                or ret.get('responses', [{}])[0].get('response_txn', {}).get('succeeded', False)
        return bool(self._lease)

    @catch_etcd_errors
    def initialize(self, create_new: bool = True, sysid: str = ""):
        return self.retry(self._client.put, self.initialize_path, sysid, create_revision='0' if create_new else None)

    @catch_etcd_errors
    def _delete_leader(self, leader: Leader) -> bool:
        fields = build_range_request(self.leader_path)
        compare = {'key': fields['key'], 'target': 'VALUE', 'value': base64_encode(self._name)}
        return bool(self._client.txn(compare, {'request_delete_range': fields}))

    @catch_etcd_errors
    def cancel_initialization(self) -> bool:
        return self.retry(self._client.deleterange, self.initialize_path)

    @catch_etcd_errors
    def delete_cluster(self) -> bool:
        return self.retry(self._client.deleteprefix, self.client_path(''))

    @catch_etcd_errors
    def set_history_value(self, value: str) -> bool:
        return bool(self._client.put(self.history_path, value))

    @catch_etcd_errors
    def set_sync_state_value(self, value: str, version: Optional[str] = None) -> Union[str, bool]:
        return self.retry(self._client.put, self.sync_path, value, mod_revision=version)\
            .get('header', {}).get('revision', False)

    @catch_etcd_errors
    def delete_sync_state(self, version: Optional[str] = None) -> bool:
        return self.retry(self._client.deleterange, self.sync_path, mod_revision=version)

    def watch(self, leader_version: Optional[str], timeout: float) -> bool:
        self._last_lease_refresh = 0
        if self.__do_not_watch:
            self.__do_not_watch = False
            return True

        # We want to give a bit more time to non-leader nodes to synchronize HA loops
        if leader_version:
            timeout += 0.5

        try:
            return super(Etcd3, self).watch(None, timeout)
        finally:
            self.event.clear()



"""Tags handling."""
import abc

from typing import Any, Dict, Optional

from patroni.utils import parse_bool, parse_int


class Tags(abc.ABC):
    """An abstract class that encapsulates all the ``tags`` logic.

    Child classes that want to use provided facilities must implement ``tags`` abstract property.

    .. note::
        Due to backward-compatibility reasons, old tags may have a less strict type conversion than new ones.
    """

    @staticmethod
    def _filter_tags(tags: Dict[str, Any]) -> Dict[str, Any]:
        """Get tags configured for this node, if any.

        Handle both predefined Patroni tags and custom defined tags.

        .. note::
            A custom tag is any tag added to the configuration ``tags`` section that is not one of ``clonefrom``,
            ``nofailover``, ``noloadbalance``,``nosync`` or ``nostream``.

            For most of the Patroni predefined tags, the returning object will only contain them if they are enabled as
            they all are boolean values that default to disabled.
            However ``nofailover`` tag is always returned if ``failover_priority`` tag is defined. In this case, we need
            both values to see if they are contradictory and the ``nofailover`` value should be used.
            The same rule applies for ``nosync`` and ``sync_priority`` tags.

        :returns: a dictionary of tags set for this node. The key is the tag name, and the value is the corresponding
            tag value.
        """
        return {tag: value for tag, value in tags.items()
                if any((tag not in ('clonefrom', 'nofailover', 'noloadbalance', 'nosync', 'nostream'),
                        value,
                        tag == 'nofailover' and 'failover_priority' in tags,
                        tag == 'nosync' and 'sync_priority' in tags))}

    @property
    @abc.abstractmethod
    def tags(self) -> Dict[str, Any]:
        """Configured tags.

        Must be implemented in a child class.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def clonefrom(self) -> bool:
        """``True`` if ``clonefrom`` tag is ``True``, else ``False``."""
        return self.tags.get('clonefrom', False)

    def _priority_tag(self, bool_name: str, priority_name: str) -> int:
        """Common logic for obtaining the value of a priority tag from ``tags`` if defined.

        If boolean tag is defined as ``True``, this will return ``0``. Otherwise, it will return the value of
        the respective priority tag, defaulting to ``1`` if it's not defined or invalid.

        :param bool_name: name of the boolean tag (``nofailover``. ``nosync``).
        :param priority_name: name of the priority tag (``failover_priority``, ``sync_priority``).

        :returns: integer value based on the defined tags.
        """
        from_tags = self.tags.get(bool_name)
        priority = parse_int(self.tags.get(priority_name))
        priority = 1 if priority is None else priority
        return 0 if from_tags else priority

    def _bool_tag(self, bool_name: str, priority_name: str) -> bool:
        """Common logic for obtaining the value of a boolean tag from ``tags`` if defined.

        If boolean tag is not defined, this methods returns ``True`` if priority tag is non-positive,
        ``False`` otherwise.

        :param bool_name: name of the boolean tag (``nofailover``. ``nosync``).
        :param priority_name: name of the priority tag (``failover_priority``, ``sync_priority``).

        :returns: boolean value based on the defined tags.
        """
        from_tags = self.tags.get(bool_name)
        if from_tags is not None:
            # Value of bool tag takes precedence over priority tag
            return bool(from_tags)
        priority = parse_int(self.tags.get(priority_name))
        return priority is not None and priority <= 0

    @property
    def nofailover(self) -> bool:
        """``True`` if node configuration doesn't allow it to become primary, ``False`` otherwise."""
        return self._bool_tag('nofailover', 'failover_priority')

    @property
    def failover_priority(self) -> int:
        """Value of ``failover_priority`` from ``tags`` if defined, otherwise derived from ``nofailover``."""
        return self._priority_tag('nofailover', 'failover_priority')

    @property
    def noloadbalance(self) -> bool:
        """``True`` if ``noloadbalance`` is ``True``, else ``False``."""
        return bool(self.tags.get('noloadbalance', False))

    @property
    def nosync(self) -> bool:
        """``True`` if node configuration doesn't allow it to become synchronous, ``False`` otherwise."""
        return self._bool_tag('nosync', 'sync_priority')

    @property
    def sync_priority(self) -> int:
        """Value of ``sync_priority`` from ``tags`` if defined, otherwise derived from ``nosync``."""
        return self._priority_tag('nosync', 'sync_priority')

    @property
    def replicatefrom(self) -> Optional[str]:
        """Value of ``replicatefrom`` tag, if any."""
        return self.tags.get('replicatefrom')

    @property
    def nostream(self) -> bool:
        """``True`` if ``nostream`` is ``True``, else ``False``."""
        return parse_bool(self.tags.get('nostream')) or False



"""Patroni main entry point.

Implement ``patroni`` main daemon and expose its entry point.
"""

import logging
import os
import signal
import sys
import time

from argparse import Namespace
from typing import Any, Dict, List, Optional, TYPE_CHECKING

from patroni import global_config, MIN_PSYCOPG2, MIN_PSYCOPG3, parse_version
from patroni.collections import EMPTY_DICT
from patroni.daemon import abstract_main, AbstractPatroniDaemon, get_base_arg_parser
from patroni.tags import Tags

if TYPE_CHECKING:  # pragma: no cover
    from .config import Config
    from .dcs import Cluster
    from .log import PatroniLogger

logger = logging.getLogger(__name__)


class Patroni(AbstractPatroniDaemon, Tags):
    """Implement ``patroni`` command daemon.

    :ivar version: Patroni version.
    :ivar dcs: DCS object.
    :ivar watchdog: watchdog handler, if configured to use watchdog.
    :ivar postgresql: managed Postgres instance.
    :ivar api: REST API server instance of this node.
    :ivar request: wrapper for performing HTTP requests.
    :ivar ha: HA handler.
    :ivar next_run: time when to run the next HA loop cycle.
    :ivar scheduled_restart: when a restart has been scheduled to occur, if any. In that case, should contain two keys:
        * ``schedule``: timestamp when restart should occur;
        * ``postmaster_start_time``: timestamp when Postgres was last started.
    """

    def __init__(self, config: 'Config', patroni_logger: 'PatroniLogger') -> None:
        """Create a :class:`Patroni` instance with the given *config*.

        Get a connection to the DCS, configure watchdog (if required), set up Patroni interface with Postgres, configure
        the HA loop and bring the REST API up.

        .. note::
            Expected to be instantiated and run through :func:`~patroni.daemon.abstract_main`.

        :param config: Patroni configuration.
        :param patroni_logger: the logging handler for this daemon.
        """
        from patroni import thread_pool
        from patroni.api import RestApiServer
        from patroni.dcs import get_dcs
        from patroni.ha import Ha
        from patroni.postgresql import Postgresql
        from patroni.request import PatroniRequest
        from patroni.version import __version__
        from patroni.watchdog import Watchdog

        try:
            thread_pool_size = max(5, int(config.get('thread_pool_size', 5)))
        except Exception as e:
            logger.warning('Failed to parse thread_pool_size value "%s": %r', config.get('thread_pool_size'), e)
            thread_pool_size = 5
        logger.info('Patroni global thread_pool_size = %d', thread_pool_size)
        thread_pool.configure_global_pool(thread_pool_size)

        super(Patroni, self).__init__(config, patroni_logger)

        self.version = __version__
        self.dcs = get_dcs(self.config)
        self.request = PatroniRequest(self.config, True)

        cluster = self.ensure_dcs_access()
        self.ensure_unique_name(cluster)

        self.watchdog = Watchdog(self.config)
        self.apply_dynamic_configuration(cluster)

        # Initialize global config
        global_config.update(None, self.config.dynamic_configuration)

        self.postgresql = Postgresql(self.config['postgresql'], self.dcs.mpp)
        self.api = RestApiServer(self, self.config['restapi'])
        self.ha = Ha(self)

        self._tags = self._get_tags()
        self.next_run = time.time()
        self.scheduled_restart: Dict[str, Any] = {}

    def ensure_dcs_access(self, sleep_time: int = 5) -> 'Cluster':
        """Continuously attempt to retrieve cluster from DCS with delay.

        :param sleep_time: seconds to wait between retry attempts after dcs connection raise :exc:`DCSError`.

        :returns: a PostgreSQL or MPP implementation of :class:`Cluster`.
        """
        from patroni.exceptions import DCSError

        while True:
            try:
                return self.dcs.get_cluster()
            except DCSError:
                logger.warning('Can not get cluster from dcs')
                time.sleep(sleep_time)

    def apply_dynamic_configuration(self, cluster: 'Cluster') -> None:
        """Apply Patroni dynamic configuration.

        Apply dynamic configuration from the DCS, if `/config` key is available in the DCS, otherwise fall back to
        ``bootstrap.dcs`` section from the configuration file.

        .. note::
            This method is called only once, at the time when Patroni is started.

        :param cluster: a PostgreSQL or MPP implementation of :class:`Cluster`.
        """
        if cluster and cluster.config and cluster.config.data:
            if self.config.set_dynamic_configuration(cluster.config):
                self.dcs.reload_config(self.config)
                self.watchdog.reload_config(self.config)
        elif not self.config.dynamic_configuration and 'bootstrap' in self.config:
            if self.config.set_dynamic_configuration(self.config['bootstrap']['dcs']):
                self.dcs.reload_config(self.config)
                self.watchdog.reload_config(self.config)

    def ensure_unique_name(self, cluster: 'Cluster') -> None:
        """A helper method to prevent splitbrain from operator naming error.

        :param cluster: a PostgreSQL or MPP implementation of :class:`Cluster`.
        """
        from patroni.dcs import Member

        if not cluster:
            return
        member = cluster.get_member(self.config['name'], False)
        if not isinstance(member, Member):
            return
        # Silence annoying WARNING: Retrying (...) messages when Patroni is quickly restarted.
        configured_loggers: Dict[str, Any] = (self.config.get('log') or EMPTY_DICT).get('loggers') or {}
        try:
            self.logger.update_loggers({**configured_loggers, 'urllib3.connectionpool': 'ERROR'})
            _ = self.request(member, endpoint="/liveness", timeout=3)
            logger.fatal("Can't start; there is already a node named '%s' running", self.config['name'])
            sys.exit(1)
        except Exception:
            self.logger.update_loggers(configured_loggers)

    def _get_tags(self) -> Dict[str, Any]:
        """Get tags configured for this node, if any.

        :returns: a dictionary of tags set for this node.
        """
        return self._filter_tags(self.config.get('tags', {}))

    def reload_config(self, sighup: bool = False, local: Optional[bool] = False) -> None:
        """Apply new configuration values for ``patroni`` daemon.

        Reload:
            * Cached tags;
            * Request wrapper configuration;
            * REST API configuration;
            * Watchdog configuration;
            * Postgres configuration;
            * DCS configuration.

        :param sighup: if it is related to a SIGHUP signal.
        :param local: if there has been changes to the local configuration file.
        """
        try:
            super(Patroni, self).reload_config(sighup, local)
            if local:
                self._tags = self._get_tags()
                self.request.reload_config(self.config)
            if local or sighup and self.api.reload_local_certificate():
                self.api.reload_config(self.config['restapi'])
            self.watchdog.reload_config(self.config)
            self.postgresql.reload_config(self.config['postgresql'], sighup)
            self.dcs.reload_config(self.config)
        except Exception:
            logger.exception('Failed to reload config_file=%s', self.config.config_file)

    @property
    def tags(self) -> Dict[str, Any]:
        """Tags configured for this node, if any."""
        return self._tags

    def schedule_next_run(self) -> None:
        """Schedule the next run of the ``patroni`` daemon main loop.

        Next run is scheduled based on previous run plus value of ``loop_wait`` configuration from DCS. If that has
        already been exceeded, run the next cycle immediately.
        """
        self.next_run += self.dcs.loop_wait
        current_time = time.time()
        nap_time = self.next_run - current_time
        if nap_time <= 0:
            self.next_run = current_time
            # Release the GIL so we don't starve anyone waiting on async_executor lock
            time.sleep(0.001)
            # Warn user that Patroni is not keeping up
            logger.warning("Loop time exceeded, rescheduling immediately.")
        elif self.ha.watch(nap_time):
            self.next_run = time.time()

    def run(self) -> None:
        """Run ``patroni`` daemon process main loop.

        Start the REST API and keep running HA cycles every ``loop_wait`` seconds.
        """
        self.api.start()
        self.next_run = time.time()
        super(Patroni, self).run()

    def _run_cycle(self) -> None:
        """Run a cycle of the ``patroni`` daemon main loop.

        Run an HA cycle and schedule the next cycle run. If any dynamic configuration change request is detected, apply
        the change and cache the new dynamic configuration values in ``patroni.dynamic.json`` file under Postgres data
        directory.
        """
        from patroni.postgresql.misc import PostgresqlRole

        logger.info(self.ha.run_cycle())

        if self.dcs.cluster and self.dcs.cluster.config and self.dcs.cluster.config.data \
                and self.config.set_dynamic_configuration(self.dcs.cluster.config):
            self.reload_config()

        if self.postgresql.role != PostgresqlRole.UNINITIALIZED:
            self.config.save_cache()

        self.schedule_next_run()

    def _shutdown(self) -> None:
        """Perform shutdown of ``patroni`` daemon process.

        Shut down the REST API and the HA handler.
        """
        from patroni import thread_pool

        try:
            self.api.shutdown()
        except Exception:
            logger.exception('Exception during RestApi.shutdown')
        try:
            self.ha.shutdown()
        except Exception:
            logger.exception('Exception during Ha.shutdown')

        thread_pool.get_executor().shutdown(wait=False)


def patroni_main(configfile: str) -> None:
    """Configure and start ``patroni`` main daemon process.

    :param configfile: path to Patroni configuration file.
    """
    abstract_main(Patroni, configfile)


def process_arguments() -> Namespace:
    """Process command-line arguments.

    Create a basic command-line parser through :func:`~patroni.daemon.get_base_arg_parser`, extend its capabilities by
    adding these flags and parse command-line arguments.:

      * ``--validate-config`` -- used to validate the Patroni configuration file
      * ``--generate-config`` -- used to generate Patroni configuration from a running PostgreSQL instance
      * ``--generate-sample-config`` -- used to generate a sample Patroni configuration
      * ``--ignore-listen-port`` | ``-i`` -- used to ignore ``listen`` ports already in use.
          Can be used only with ``--validate-config``
      * ``--print`` | ``-p`` -- used to print out local configuration (incl. environment configuration overrides).
          Can be used only with ``--validate-config``

    .. note::
        If running with ``--generate-config``, ``--generate-sample-config`` or ``--validate-flag`` will exit
        after generating or validating configuration.

    :returns: parsed arguments, if not running with ``--validate-config`` flag.
    """
    from patroni.config_generator import generate_config

    parser = get_base_arg_parser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--validate-config', action='store_true', help='Run config validator and exit')
    group.add_argument('--generate-sample-config', action='store_true',
                       help='Generate a sample Patroni yaml configuration file')
    group.add_argument('--generate-config', action='store_true',
                       help='Generate a Patroni yaml configuration file for a running instance')
    parser.add_argument('--dsn', help='Optional DSN string of the instance to be used as a source \
                                    for config generation. Superuser connection is required.')
    parser.add_argument('--ignore-listen-port', '-i', action='store_true',
                        help='Ignore `listen` ports already in use.\
                              Can only be used with --validate-config')
    parser.add_argument('--print', '-p', action='store_true',
                        help='Print out local configuration (incl. environment configuration overrides).\
                              Can only be used with --validate-config')
    args = parser.parse_args()

    if args.generate_sample_config:
        generate_config(args.configfile, True, None)
        sys.exit(0)
    elif args.generate_config:
        generate_config(args.configfile, False, args.dsn)
        sys.exit(0)
    elif args.validate_config:
        from patroni.config import Config, ConfigParseError
        from patroni.validator import populate_validate_params, schema

        populate_validate_params(ignore_listen_port=args.ignore_listen_port)

        try:
            config = Config(args.configfile, validator=schema)
        except ConfigParseError as e:
            sys.exit(e.value)

        if args.print:
            import yaml
            yaml.safe_dump(config.local_configuration, sys.stdout, default_flow_style=False, allow_unicode=True)
        sys.exit()

    return args


def check_psycopg() -> None:
    """Ensure at least one among :mod:`psycopg2` or :mod:`psycopg` libraries are available in the environment.

    .. note::
        Patroni chooses :mod:`psycopg2` over :mod:`psycopg`, if possible.

        If nothing meeting the requirements is found, then exit with a fatal message.
    """
    min_psycopg2_str = '.'.join(map(str, MIN_PSYCOPG2))
    min_psycopg3_str = '.'.join(map(str, MIN_PSYCOPG3))

    available_versions: List[str] = []

    # try psycopg2
    try:
        from psycopg2 import __version__
        if parse_version(__version__) >= MIN_PSYCOPG2:
            return
        available_versions.append('psycopg2=={0}'.format(__version__.split(' ')[0]))
    except ImportError:
        logger.debug('psycopg2 module is not available')

    # try psycopg3
    try:
        from psycopg import __version__
        if parse_version(__version__) >= MIN_PSYCOPG3:
            return
        available_versions.append('psycopg=={0}'.format(__version__.split(' ')[0]))
    except ImportError:
        logger.debug('psycopg module is not available')

    error = f'FATAL: Patroni requires psycopg2>={min_psycopg2_str}, psycopg2-binary, or psycopg>={min_psycopg3_str}'
    if available_versions:
        error += ', but only {0} {1} available'.format(
            ' and '.join(available_versions),
            'is' if len(available_versions) == 1 else 'are')
    sys.exit(error)


def main() -> None:
    """Main entrypoint of :mod:`patroni.__main__`.

    Process command-line arguments, ensure :mod:`psycopg2` (or :mod:`psycopg`) attendee the pre-requisites and start
    ``patroni`` daemon process.

    .. note::
        If running through a Docker container, make the main process take care of init process duties and run
        ``patroni`` daemon as another process. In that case relevant signals received by the main process and forwarded
        to ``patroni`` daemon process.
    """
    from multiprocessing import freeze_support

    # Executables created by PyInstaller are frozen, thus we need to enable frozen support for
    # :mod:`multiprocessing` to avoid :class:`RuntimeError` exceptions.
    freeze_support()

    check_psycopg()

    args = process_arguments()

    if os.getpid() != 1:
        return patroni_main(args.configfile)

    # Patroni started with PID=1, it looks like we are in the container
    from types import FrameType
    pid = 0

    # Looks like we are in a docker, so we will act like init
    def sigchld_handler(signo: int, stack_frame: Optional[FrameType]) -> None:
        """Handle ``SIGCHLD`` received by main process from ``patroni`` daemon when the daemon terminates.

        :param signo: signal number.
        :param stack_frame: current stack frame.
        """
        try:
            # log exit code of all children processes, and break loop when there is none left
            while True:
                ret = os.waitpid(-1, os.WNOHANG)
                if ret == (0, 0):
                    break
                elif ret[0] != pid:
                    logger.info('Reaped pid=%s, exit status=%s', *ret)
        except OSError:
            pass

    def passtochild(signo: int, stack_frame: Optional[FrameType]) -> None:
        """Forward a signal *signo* from main process to child process.

        :param signo: signal number.
        :param stack_frame: current stack frame.
        """
        if pid:
            os.kill(pid, signo)

    import multiprocessing
    patroni = multiprocessing.Process(target=patroni_main, args=(args.configfile,))
    patroni.start()
    pid = patroni.pid

    # Set up signal handlers after fork to prevent child from inheriting them
    if os.name != 'nt':
        signal.signal(signal.SIGCHLD, sigchld_handler)
        signal.signal(signal.SIGHUP, passtochild)
        signal.signal(signal.SIGQUIT, passtochild)
        signal.signal(signal.SIGUSR1, passtochild)
        signal.signal(signal.SIGUSR2, passtochild)
    signal.signal(signal.SIGINT, passtochild)
    signal.signal(signal.SIGABRT, passtochild)
    signal.signal(signal.SIGTERM, passtochild)
    patroni.join()


if __name__ == '__main__':
    main()



import json
import logging
import random
import time

from typing import Any, Callable, cast, Dict, List, Union

from ..postgresql.mpp import AbstractMPP
from ..request import get as requests_get
from ..utils import uri
from . import Cluster
from .zookeeper import ZooKeeper

logger = logging.getLogger(__name__)


class ExhibitorEnsembleProvider(object):

    TIMEOUT = 3.1

    def __init__(self, hosts: List[str], port: int,
                 uri_path: str = '/exhibitor/v1/cluster/list', poll_interval: int = 300) -> None:
        self._exhibitor_port = port
        self._uri_path = uri_path
        self._poll_interval = poll_interval
        self._exhibitors: List[str] = hosts
        self._boot_exhibitors = hosts
        self._zookeeper_hosts = ''
        self._next_poll = None
        while not self.poll():
            logger.info('waiting on exhibitor')
            time.sleep(5)

    def poll(self) -> bool:
        if self._next_poll and self._next_poll > time.time():
            return False

        json = self._query_exhibitors(self._exhibitors)
        if not json:
            json = self._query_exhibitors(self._boot_exhibitors)

        if isinstance(json, dict) and 'servers' in json and 'port' in json:
            self._next_poll = time.time() + self._poll_interval
            servers: List[str] = cast(Dict[str, Any], json)['servers']
            port = str(cast(Dict[str, Any], json)['port'])
            zookeeper_hosts = ','.join([h + ':' + port for h in sorted(servers)])
            if self._zookeeper_hosts != zookeeper_hosts:
                logger.info('ZooKeeper connection string has changed: %s => %s', self._zookeeper_hosts, zookeeper_hosts)
                self._zookeeper_hosts = zookeeper_hosts
                self._exhibitors = json['servers']
                return True
        return False

    def _query_exhibitors(self, exhibitors: List[str]) -> Any:
        random.shuffle(exhibitors)
        for host in exhibitors:
            try:
                response = requests_get(uri('http', (host, self._exhibitor_port), self._uri_path), timeout=self.TIMEOUT)
                return json.loads(response.data.decode('utf-8'))
            except Exception:
                logging.debug('Request to %s failed', host)
        return None

    @property
    def zookeeper_hosts(self) -> str:
        return self._zookeeper_hosts


class Exhibitor(ZooKeeper):

    def __init__(self, config: Dict[str, Any], mpp: AbstractMPP) -> None:
        interval = config.get('poll_interval', 300)
        self._ensemble_provider = ExhibitorEnsembleProvider(config['hosts'], config['port'], poll_interval=interval)
        super(Exhibitor, self).__init__({**config, 'hosts': self._ensemble_provider.zookeeper_hosts}, mpp)

    def _load_cluster(
            self, path: str, loader: Callable[[str], Union[Cluster, Dict[int, Cluster]]]
    ) -> Union[Cluster, Dict[int, Cluster]]:
        if self._ensemble_provider.poll():
            self._client.set_hosts(self._ensemble_provider.zookeeper_hosts)
        return super(Exhibitor, self)._load_cluster(path, loader)



"""Patroni logging facilities.

Daemon processes will use a 2-step logging handler. Whenever a log message is issued it is initially enqueued in-memory
and is later asynchronously flushed by a thread to the final destination.
"""
import logging
import os
import sys

from copy import deepcopy
from io import TextIOWrapper
from logging.handlers import RotatingFileHandler
from queue import Full, Queue
from threading import Lock, Thread
from typing import Any, cast, Dict, List, Optional, TYPE_CHECKING, Union

from .file_perm import pg_perm
from .utils import deep_compare, parse_int

type_logformat = Union[List[Union[str, Dict[str, Any], Any]], str, Any]

_LOGGER = logging.getLogger(__name__)


class PatroniFileHandler(RotatingFileHandler):
    """Wrapper of :class:`RotatingFileHandler` to handle permissions of log files. """

    def __init__(self, filename: str, mode: Optional[int]) -> None:
        """Create a new :class:`PatroniFileHandler` instance.

        :param filename: basename for log files.
        :param mode: permissions for log files.
        """
        self.set_log_file_mode(mode)
        super(PatroniFileHandler, self).__init__(filename)

    def set_log_file_mode(self, mode: Optional[int]) -> None:
        """Set mode for Patroni log files.

        :param mode: permissions for log files.

        .. note::
            If *mode* is not specified, we calculate it from the `umask` value.
        """
        self._log_file_mode = 0o666 & ~pg_perm.orig_umask if mode is None else mode

    def _open(self) -> TextIOWrapper:
        """Open a new log file and assign permissions.

        :returns: the resulting stream.
        """
        ret = super(PatroniFileHandler, self)._open()
        os.chmod(self.baseFilename, self._log_file_mode)
        return ret


def debug_exception(self: logging.Logger, msg: object, *args: Any, **kwargs: Any) -> None:
    """Add full stack trace info to debug log messages and partial to others.

    Handle :func:`~self.exception` calls for *self*.

    .. note::
        * If *self* log level is set to ``DEBUG``, then issue a ``DEBUG`` message with the complete stack trace;
        * If *self* log level is ``INFO`` or higher, then issue an ``ERROR`` message with only the last line of
            the stack trace.

    :param self: logger for which :func:`~self.exception` will be processed.
    :param msg: the message related to the exception to be logged.
    :param args: positional arguments to be passed to :func:`~self.debug` or :func:`~self.error`.
    :param kwargs: keyword arguments to be passed to :func:`~self.debug` or :func:`~self.error`.
    """
    kwargs.pop("exc_info", False)
    if self.isEnabledFor(logging.DEBUG):
        self.debug(msg, *args, exc_info=True, **kwargs)
    else:
        msg = "{0}, DETAIL: '{1}'".format(msg, sys.exc_info()[1])
        self.error(msg, *args, exc_info=False, **kwargs)


def error_exception(self: logging.Logger, msg: object, *args: Any, **kwargs: Any) -> None:
    """Add full stack trace info to error messages.

    Handle :func:`~self.exception` calls for *self*.

    .. note::
        * By default issue an ``ERROR`` message with the complete stack trace. If you do not want to show the complete
          stack trace, call with ``exc_info=False``.

    :param self: logger for which :func:`~self.exception` will be processed.
    :param msg: the message related to the exception to be logged.
    :param args: positional arguments to be passed to :func:`~self.error`.
    :param kwargs: keyword arguments to be passed to :func:`~self.error`.
    """
    exc_info = kwargs.pop("exc_info", True)
    self.error(msg, *args, exc_info=exc_info, **kwargs)


def _type(value: Any) -> str:
    """Get type of the *value*.

    :param value: any arbitrary value.
    :returns: a string with a type name.
    """
    return value.__class__.__name__


class QueueHandler(logging.Handler):
    """Queue-based logging handler.

    :ivar queue: queue to hold log messages that are pending to be flushed to the final destination.
    """

    def __init__(self) -> None:
        """Queue initialised and initial records_lost established."""
        super().__init__()
        self.queue: Queue[Union[logging.LogRecord, None]] = Queue()
        self._records_lost = 0

    def _put_record(self, record: logging.LogRecord) -> None:
        """Asynchronously enqueue a log record.

        :param record: the record to be logged.
        """
        self.format(record)
        record.msg = record.message
        record.args = None
        record.exc_info = None
        self.queue.put_nowait(record)

    def _try_to_report_lost_records(self) -> None:
        """Report the number of log messages that have been lost and reset the counter.

        .. note::
            It will issue an ``WARNING`` message in the logs with the number of lost log messages.
        """
        if self._records_lost:
            try:
                record = _LOGGER.makeRecord(_LOGGER.name, logging.WARNING, __file__, 0,
                                            'QueueHandler has lost %s log records',
                                            (self._records_lost,), None, 'emit')
                self._put_record(record)
                self._records_lost = 0
            except Exception:
                pass

    def emit(self, record: logging.LogRecord) -> None:
        """Handle each log record that is emitted.

        Call :func:`_put_record` to enqueue the emitted log record.

        Also check if we have previously lost any log record, and if so, log a ``WARNING`` message.

        :param record: the record that was emitted.
        """
        try:
            self._put_record(record)
            self._try_to_report_lost_records()
        except Exception:
            self._records_lost += 1

    @property
    def records_lost(self) -> int:
        """Number of log messages that have been lost while the queue was full."""
        return self._records_lost


class ProxyHandler(logging.Handler):
    """Handle log records in place of pending log handlers.

    .. note::
        This is used to handle log messages while the logger thread has not started yet, in which case the queue-based
        handler is not yet started.

    :ivar patroni_logger: the logger thread.
    """

    def __init__(self, patroni_logger: 'PatroniLogger') -> None:
        """Create a new :class:`ProxyHandler` instance.

        :param patroni_logger: the logger thread.
        """
        super().__init__()
        self.patroni_logger = patroni_logger

    def emit(self, record: logging.LogRecord) -> None:
        """Emit each log record that is handled.

        Will push the log record down to :func:`~logging.Handler.handle` method of the currently configured log handler.

        :param record: the record that was emitted.
        """
        if self.patroni_logger.log_handler is not None:
            self.patroni_logger.log_handler.handle(record)


class PatroniLogger(Thread):
    """Logging thread for the Patroni daemon process.

    It is a 2-step logging approach. Any time a log message is issued it is initially enqueued in-memory, and then
    asynchronously flushed to the final destination by the logging thread.

    .. seealso::
        :class:`QueueHandler`: object used for enqueueing messages in-memory.

    :cvar DEFAULT_TYPE: default type of log format (``plain``).
    :cvar DEFAULT_LEVEL: default logging level (``INFO``).
    :cvar DEFAULT_TRACEBACK_LEVEL: default traceback logging level (``ERROR``).
    :cvar DEFAULT_FORMAT: default format of log messages (``%(asctime)s %(levelname)s: %(message)s``).
    :cvar NORMAL_LOG_QUEUE_SIZE: expected number of log messages per HA loop when operating under a normal situation.
    :cvar DEFAULT_MAX_QUEUE_SIZE: default maximum queue size for holding a backlog of log messages that are pending
        to be flushed.
    :cvar LOGGING_BROKEN_EXIT_CODE: exit code to be used if it detects(``5``).

    :ivar log_handler: log handler that is currently being used by the thread.
    :ivar log_handler_lock: lock used to modify ``log_handler``.
    """

    DEFAULT_TYPE = 'plain'
    DEFAULT_LEVEL = 'INFO'
    DEFAULT_TRACEBACK_LEVEL = 'ERROR'
    DEFAULT_FORMAT = '%(asctime)s %(levelname)s: %(message)s'

    NORMAL_LOG_QUEUE_SIZE = 2  # When everything goes normal Patroni writes only 2 messages per HA loop
    DEFAULT_MAX_QUEUE_SIZE = 1000
    LOGGING_BROKEN_EXIT_CODE = 5

    def __init__(self) -> None:
        """Prepare logging queue and proxy handlers as they become ready during daemon startup.

        .. note::
            While Patroni is starting up it keeps ``DEBUG`` log level, and writes log messages through a proxy handler.
            Once the logger thread is finally started, it switches from that proxy handler to the queue based logger,
            and applies the configured log settings. The switching is used to avoid that the logger thread prevents
            Patroni from shutting down if any issue occurs in the meantime until the thread is properly started.
        """
        super(PatroniLogger, self).__init__()
        self._queue_handler = QueueHandler()
        self._root_logger = logging.getLogger()
        self._config: Optional[Dict[str, Any]] = None
        self.log_handler = None
        self.log_handler_lock = Lock()
        self._old_handlers: List[logging.Handler] = []
        # initially set log level to ``DEBUG`` while the logger thread has not started running yet. The daemon process
        # will later adjust all log related settings with what was provided through the user configuration file.
        self.reload_config({'level': 'DEBUG'})
        # We will switch to the QueueHandler only when thread was started.
        # This is necessary to protect from the cases when Patroni constructor
        # failed and PatroniLogger thread remain running and prevent shutdown.
        self._proxy_handler = ProxyHandler(self)
        self._root_logger.addHandler(self._proxy_handler)

    def update_loggers(self, config: Dict[str, Any]) -> None:
        """Configure custom loggers' log levels.

        .. note::
            It creates logger objects that are not defined yet in the log manager.

        :param config: :class:`dict` object with custom loggers configuration, is set either from:

                       * ``log.loggers`` section of Patroni configuration; or

                       * from the method that is trying to make sure that the node name
                         isn't duplicated (to silence annoying ``urllib3`` WARNING's).

        :Example:

            .. code-block:: python

                update_loggers({'urllib3.connectionpool': 'WARNING'})
        """
        loggers = deepcopy(config)
        for name, logger in self._root_logger.manager.loggerDict.items():
            # ``Placeholder`` is a node in the log manager for which no logger has been defined. We are interested only
            # in the ones that were defined
            if not isinstance(logger, logging.PlaceHolder):
                # if this logger is present in *config*, use the configured level, otherwise
                # use ``logging.NOTSET``, which means it will inherit the level
                # from any parent node up to the root for which log level is defined.
                level = loggers.pop(name, logging.NOTSET)
                logger.setLevel(level)

        # define loggers that do not exist yet and set level as configured in the *config*
        for name, level in loggers.items():
            logger = self._root_logger.manager.getLogger(name)
            logger.setLevel(level)

    def _is_config_changed(self, config: Dict[str, Any]) -> bool:
        """Checks if the given config is different from the current one.

        :param config: ``log`` section from Patroni configuration.

        :returns: ``True`` if the config is changed, ``False`` otherwise.
        """
        old_config = self._config or {}

        oldlogtype = old_config.get('type', PatroniLogger.DEFAULT_TYPE)
        logtype = config.get('type', PatroniLogger.DEFAULT_TYPE)

        oldlogformat: type_logformat = old_config.get('format', PatroniLogger.DEFAULT_FORMAT)
        logformat: type_logformat = config.get('format', PatroniLogger.DEFAULT_FORMAT)

        olddateformat = old_config.get('dateformat') or None
        dateformat = config.get('dateformat') or None  # Convert empty string to `None`

        old_static_fields = old_config.get('static_fields', {})
        static_fields = config.get('static_fields', {})

        old_log_config = {
            'type': oldlogtype,
            'format': oldlogformat,
            'dateformat': olddateformat,
            'static_fields': old_static_fields
        }

        log_config = {
            'type': logtype,
            'format': logformat,
            'dateformat': dateformat,
            'static_fields': static_fields
        }

        return not deep_compare(old_log_config, log_config)

    def _get_plain_formatter(self, logformat: type_logformat, dateformat: Optional[str]) -> logging.Formatter:
        """Returns a logging formatter with the specified format and date format.

        .. note::
            If the log format isn't a string, prints a warning message and uses the default log format instead.

        :param logformat: The format of the log messages.
        :param dateformat: The format of the timestamp in the log messages.

        :returns: A logging formatter object that can be used to format log records.
        """

        if not isinstance(logformat, str):
            _LOGGER.warning('Expected log format to be a string when log type is plain, but got "%s"', _type(logformat))
            logformat = PatroniLogger.DEFAULT_FORMAT

        return logging.Formatter(logformat, dateformat)

    def _get_json_formatter(self, logformat: type_logformat, dateformat: Optional[str],
                            static_fields: Dict[str, Any]) -> logging.Formatter:
        """Returns a logging formatter that outputs JSON formatted messages.

        .. note::
            If :mod:`pythonjsonlogger` library is not installed, prints an error message and returns
            a plain log formatter instead.

        :param logformat: Specifies the log fields and their key names in the JSON log message.
        :param dateformat: The format of the timestamp in the log messages.
        :param static_fields: A dictionary of static fields that are added to every log message.

        :returns: A logging formatter object that can be used to format log records as JSON strings.
        """

        if isinstance(logformat, str):
            jsonformat = logformat
            rename_fields = {}
        elif isinstance(logformat, list):
            logformat = cast(List[Any], logformat)
            log_fields: List[str] = []
            rename_fields: Dict[str, str] = {}

            for field in logformat:
                if isinstance(field, str):
                    log_fields.append(field)
                elif isinstance(field, dict):
                    field = cast(Dict[str, Any], field)
                    for original_field, renamed_field in field.items():
                        if isinstance(renamed_field, str):
                            log_fields.append(original_field)
                            rename_fields[original_field] = renamed_field
                        else:
                            _LOGGER.warning(
                                'Expected renamed log field to be a string, but got "%s"',
                                _type(renamed_field)
                            )

                else:
                    _LOGGER.warning(
                        'Expected each item of log format to be a string or dictionary, but got "%s"',
                        _type(field)
                    )

            if len(log_fields) > 0:
                jsonformat = ' '.join([f'%({field})s' for field in log_fields])
            else:
                jsonformat = PatroniLogger.DEFAULT_FORMAT
        else:
            jsonformat = PatroniLogger.DEFAULT_FORMAT
            rename_fields = {}
            _LOGGER.warning('Expected log format to be a string or a list, but got "%s"', _type(logformat))

        try:
            try:
                from pythonjsonlogger import json as jsonlogger  # pyright: ignore
            except ImportError:  # pragma: no cover
                from pythonjsonlogger import jsonlogger
                if hasattr(jsonlogger, 'RESERVED_ATTRS') \
                        and 'taskName' not in jsonlogger.RESERVED_ATTRS:  # pyright: ignore [reportPrivateImportUsage]
                    # compatibility with python 3.12, that added a new attribute to LogRecord
                    jsonlogger.RESERVED_ATTRS += ('taskName',)  # pyright: ignore

            return jsonlogger.JsonFormatter(  # pyright: ignore [reportPrivateImportUsage]
                jsonformat,
                dateformat,
                rename_fields=rename_fields,
                static_fields=static_fields
            )
        except ImportError as e:
            _LOGGER.error('Failed to import "python-json-logger" library: %r. Falling back to the plain logger', e)
        except Exception as e:
            _LOGGER.error('Failed to initialize JsonFormatter: %r. Falling back to the plain logger', e)

        return self._get_plain_formatter(jsonformat, dateformat)

    def _get_formatter(self, config: Dict[str, Any]) -> logging.Formatter:
        """Returns a logging formatter based on the type of logger in the given configuration.

        :param config: ``log`` section from Patroni configuration.

        :returns: A :class:`logging.Formatter` object that can be used to format log records.
        """
        logtype = config.get('type', PatroniLogger.DEFAULT_TYPE)
        logformat: type_logformat = config.get('format', PatroniLogger.DEFAULT_FORMAT)
        dateformat = config.get('dateformat') or None  # Convert empty string to `None`
        static_fields = config.get('static_fields', {})

        if dateformat is not None and not isinstance(dateformat, str):
            _LOGGER.warning('Expected log dateformat to be a string, but got "%s"', _type(dateformat))
            dateformat = None

        if logtype == 'json':
            formatter = self._get_json_formatter(logformat, dateformat, static_fields)
        else:
            formatter = self._get_plain_formatter(logformat, dateformat)

        return formatter

    def reload_config(self, config: Dict[str, Any]) -> None:
        """Apply log related configuration.

        .. note::
            It is also able to deal with runtime configuration changes.

        :param config: ``log`` section from Patroni configuration.
        """
        if self._config is None or not deep_compare(self._config, config):
            with self._queue_handler.queue.mutex:
                self._queue_handler.queue.maxsize = config.get('max_queue_size', self.DEFAULT_MAX_QUEUE_SIZE)

            self._root_logger.setLevel(config.get('level', PatroniLogger.DEFAULT_LEVEL))
            if config.get('traceback_level', PatroniLogger.DEFAULT_TRACEBACK_LEVEL).lower() == 'debug':
                # show stack traces only if ``log.traceback_level`` is ``DEBUG``
                logging.Logger.exception = debug_exception
            else:
                # show stack traces as ``ERROR`` log messages
                logging.Logger.exception = error_exception

            handler = self.log_handler

            if 'dir' in config:
                mode = parse_int(config.get('mode'))
                if not isinstance(handler, PatroniFileHandler):
                    handler = PatroniFileHandler(os.path.join(config['dir'], __name__), mode)
                handler.set_log_file_mode(mode)
                max_file_size = int(config.get('file_size', 25000000))
                handler.maxBytes = max_file_size  # pyright: ignore [reportAttributeAccessIssue]
                handler.backupCount = int(config.get('file_num', 4))
            # we can't use `if not isinstance(handler, logging.StreamHandler)` below,
            # because RotatingFileHandler and PatroniFileHandler are children of StreamHandler!!!
            elif handler is None or isinstance(handler, PatroniFileHandler):
                handler = logging.StreamHandler()

            is_new_handler = handler != self.log_handler

            if (self._is_config_changed(config) or is_new_handler) and handler:
                formatter = self._get_formatter(config)
                handler.setFormatter(formatter)

            if is_new_handler:
                with self.log_handler_lock:
                    if self.log_handler:
                        self._old_handlers.append(self.log_handler)
                    self.log_handler = handler

            self._config = config.copy()
            self.update_loggers(config.get('loggers') or {})

    def _close_old_handlers(self) -> None:
        """Close old log handlers.

        .. note::
            It is used to remove different handlers that were configured previous to a reload in the configuration,
            e.g. if we are switching from :class:`PatroniFileHandler` to
            class:`~logging.StreamHandler` and vice-versa.
        """
        while True:
            with self.log_handler_lock:
                if not self._old_handlers:
                    break
                handler = self._old_handlers.pop()
            try:
                handler.close()
            except Exception:
                _LOGGER.exception('Failed to close the old log handler %s', handler)

    def run(self) -> None:
        """Run logger's thread main loop.

        Keep consuming log queue until requested to quit through ``None`` special log record.
        """
        # switch to QueueHandler only when the thread was started
        with self.log_handler_lock:
            self._root_logger.addHandler(self._queue_handler)
            self._root_logger.removeHandler(self._proxy_handler)

        prev_record = None
        prev_hb_msg = ''

        while True:
            self._close_old_handlers()
            if TYPE_CHECKING:  # pragma: no cover
                assert self.log_handler is not None

            record = self._queue_handler.queue.get(True)
            # special message that indicates Patroni is shutting down
            if record is None:
                break

            if self._root_logger.level == logging.INFO:
                # messages like ``Lock owner: postgresql0; I am postgresql1`` will be shown only when stream doesn't
                # look normal. This is used to reduce chattiness of Patroni logs.
                if record.msg.startswith('Lock owner: '):
                    prev_record, record = record, None
                else:
                    if prev_record and prev_record.thread == record.thread:
                        if self._is_heartbeat_msg(record):
                            config = self._config or {}
                            deduplicate_heartbeat_logs = config.get('deduplicate_heartbeat_logs', False)
                            if record.msg == prev_hb_msg and deduplicate_heartbeat_logs:
                                record = None
                            else:
                                prev_hb_msg = record.msg
                        else:
                            self.log_handler.handle(prev_record)
                            prev_hb_msg = None
                        prev_record = None

            if record:
                self.log_handler.handle(record)

            self._queue_handler.queue.task_done()

    @staticmethod
    def _is_heartbeat_msg(record: logging.LogRecord) -> bool:
        """Checks if the given record contains a heartbeat message.

        :param record: the record to check.

        :returns: ``True`` if the record contains a heartbeat message, ``False`` otherwise.
        """
        return record.msg.startswith('no action. ') or record.msg.startswith('PAUSE: no action')

    def shutdown(self) -> None:
        """Shut down the logger thread."""
        try:
            # ``None`` is a special message indicating to queue handler that it should quit its main loop.
            self._queue_handler.queue.put_nowait(None)
        except Full:  # Queue is full.
            # It seems that logging is not working, exiting with non-standard exit-code is the best we can do.
            sys.exit(self.LOGGING_BROKEN_EXIT_CODE)
        self.join()
        logging.shutdown()

    @property
    def queue_size(self) -> int:
        """Number of log records in the queue."""
        return self._queue_handler.queue.qsize()

    @property
    def records_lost(self) -> int:
        """Number of logging records that have been lost while the queue was full."""
        return self._queue_handler.records_lost
